# system modules
import functools
import itertools
import sys
import datetime
import logging
import inspect
import operator
import re

# internal modules
from polt.l10n import _
from polt.animator.lines import LinesAnimator
from polt.utils import *

# external modules
import numpy as np
import scipy.interpolate
import scipy.signal
import scipy.fftpack
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)


class SpectrumAnimator(LinesAnimator):
    """
    Animator to display recieved data as lines.

    Args:
        args, kwargs: further arguments handed to the
            :any:`LinesAnimator` constructor.
        blocks (int, optional): Into how many block-wise averages should be
            taken of the FFT. Default is to not split the data up.
        hanning (bool, optional): Whether to apply a
            :any:`numpy.hanning`-window before the FFT. Default is ``True``.
        time_frame (int, optional): displayed time frame in seconds. By default
            (``None``), all times are diplayed.
    """

    def __init__(self, *args, time_frame=None, hanning=None, **kwargs):
        LinesAnimator.__init__(self, *args, **kwargs)
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @property
    def hanning(self):
        """
        Whether to apply a :any:`numpy.hanning`-window before the FFT.

        .. note::

            Set this option from the command-line via
            ``polt live -a spectrum -o hanning=yes|no``

        :type: :class:`bool`
        """
        try:
            self._hanning
        except AttributeError:
            self._hanning = False
        return self._hanning

    @hanning.setter
    def hanning(self, new):
        self._hanning = new == "yes" if isinstance(new, str) else bool(new)

    @property
    def normed(self):
        """
        Whether to norm the spectrum with the frequency. Default is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -a spectrum -o normed=yes|no``

        :type: :class:`str` or ``None``
        """
        try:
            self._normed
        except AttributeError:
            self._normed = False
        return self._normed

    @normed.setter
    def normed(self, new):
        self._normed = new == "yes" if isinstance(new, str) else bool(new)

    @property
    def detrend(self):
        """
        Whether to detrend the data before preforming the FFT.

        .. note::

            Set this option from the command-line via
            ``polt live -a spectrum -o detrend=yes|no``

        :type: :class:`str` or ``None``
        """
        try:
            self._detrend
        except AttributeError:
            self._detrend = True
        return self._detrend

    @detrend.setter
    def detrend(self, new):
        self._detrend = new == "yes" if isinstance(new, str) else bool(new)

    @property
    def blocks(self):
        """
        Into how many blocks to split the data for later averaging of the FFT.
        Default is to not split the data up.

        .. note::

            Set this option from the command-line via
            ``polt live -a spectrum -o blocks=N``

        :type: :class:`bool`
        """
        try:
            self._blocks
        except AttributeError:
            self._blocks = 1
        return self._blocks

    @blocks.setter
    def blocks(self, new):
        self._blocks = max(1, int(new))

    @property
    def time_frame(self):
        """
        Time frame to use in seconds. Data outside that window
        will be dropped.

        .. note::

            Set this option from the command-line via
            ``polt live -a spectrum -o time-frame=SECONDS``

        :type: :class:`int` or ``None``
        """
        try:
            self._time_frame
        except AttributeError:
            self._time_frame = None
        return self._time_frame

    @time_frame.setter
    def time_frame(self, new):
        self._time_frame = max(sys.float_info.min, int(new))

    @property
    def show_data_rate(self):
        """
        Whether to show the displayed data rate in the legend label. The
        default is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -a spectrum -o show-data-rate=yes|no``

        :type: :class:`bool`
        """
        try:
            self._show_data_rate
        except AttributeError:
            self._show_data_rate = False
        return self._show_data_rate

    @show_data_rate.setter
    def show_data_rate(self, new):
        self._show_data_rate = (
            new == "yes" if isinstance(new, str) else bool(new)
        )

    def spectrum(
        self, x, y, hanning=None, blocks=None, normed=None, detrend=None
    ):
        """
        Call :any:`numpy.fft.fft` and :any:`numpy.fft.fftfreq` with default
        arguments.

        Args:
            x,y (sequences of floats): x and y values
            hanning (bool, optional): whether to apply a Hanning window
            normed (bool, optional): whether to norm the power spectrum with
                the frequency.
            detrend (bool, optional): whether to detrend the data first

        Returns:
            sequence : frequencies, power as arrays
        """
        hanning = self.hanning if hanning is None else hanning
        blocks = self.blocks if blocks is None else blocks
        normed = self.normed if normed is None else normed
        detrend = self.detrend if detrend is None else detrend
        # convert inputs to numpy arrays
        x, y = map(np.asarray, (x, y))
        # drop older values to fit the number of blocks
        if x.size > blocks:
            drop = x.size % blocks
            x, y = x[drop:], y[drop:]
        # interpolate to evenly-spaced times
        if x.size > 1:
            interpolator = scipy.interpolate.interp1d(x, y)
            x = np.linspace(x.min(), x.max(), num=x.size)
            y = interpolator(x)
        # detrend
        if detrend:
            y = scipy.signal.detrend(y)
        # calculate power spectrum
        power = (
            np.abs(
                # calculate average across block means
                np.mean(
                    # stack fft block resuls
                    np.vstack(
                        tuple(
                            # Fourrier transformation
                            map(
                                scipy.fftpack.fft,
                                # apply hanning window if desired
                                map(
                                    lambda a: a * np.hanning(a.size)
                                    if hanning
                                    else a,
                                    # split y into blocks
                                    np.split(y, blocks)
                                    if y.size > blocks
                                    else (y,),
                                ),
                            )
                        )
                    ),
                    axis=0,
                )
            )
            ** 2
        )
        # calculate frequencies
        freq = np.fft.fftfreq(power.size, ((x.max() - x.min()) / x.size) or 1)
        # sort frequencies
        idx = np.argsort(freq)
        freq, power = freq[idx], power[idx]
        positive_freq = freq >= 0
        return (
            freq[positive_freq],
            power[positive_freq] / (freq[positive_freq] if normed else 1),
        )

    def new_line_coords(self, values, dataset={}):
        new_value = np.mean(np.array(values))
        new_time = np.datetime64(
            dataset.get("time_recieved_utc", datetime.datetime.utcnow())
        )
        times, values = new_time.reshape(-1), new_value.reshape(-1)
        freq, fft = self.spectrum(
            (times - times.min()) / np.timedelta64(1, "s"), values
        )
        return freq, fft, {}, {"times": times, "values": values}

    def update_line(self, line, values, dataset={}):
        line.values = np.append(line.values, values)
        line.times = np.append(
            line.times,
            np.datetime64(
                dataset.get("time_recieved_utc", datetime.datetime.utcnow())
            ),
        )
        freq, fft = self.spectrum(
            (line.times - line.times.min()) / np.timedelta64(1, "s"),
            line.values,
        )
        line.set_xdata(freq[1:] if self.detrend else freq)
        line.set_ydata(fft[1:] if self.detrend else fft)

    @LinesAnimator.register_method("key-pressed")
    def toggle_normed(self, event):
        """
        Callback to toggle :any:`normed` if the pressed
        key was an ``n``.

        .. hint::

            This means, you can press ``n`` when using the
            :any:`SpectrumAnimator` to switch between normed and unnormed
            spectra.

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == "n":
            self.normed = not self.normed
            for ax in self.figure.axes:
                self.update_ylabel(ax)

    @LinesAnimator.register_method("key-pressed")
    def toggle_detrend(self, event):
        """
        Callback to toggle :any:`detrend` if the pressed
        key was an ``d``.

        .. hint::

            This means, you can press ``d`` when using the
            :any:`SpectrumAnimator` to toggle detrending.

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == "d":
            self.detrend = not self.detrend
            for ax in self.figure.axes:
                self.update_ylabel(ax)

    @LinesAnimator.register_method("before-line-update")
    def drop_old_values(self, line):
        if self.time_frame:
            outside = (
                np.datetime64(datetime.datetime.utcnow()) - line.times
            ) / np.timedelta64(1, "s") > self.time_frame
            if outside.any():
                outside_ind = np.where(outside)[0]
                line.values = np.delete(line.values, outside_ind)
                line.times = np.delete(line.times, outside_ind)

    @LinesAnimator.register_method("after-each-frame")
    def adjust_view(self, *args, **kwargs):
        """
        Adjust the view with :any:`LinesAnimator.fit_view` for both axes.
        """
        self.fit_view(axes="xy")

    def update_ylabel(self, ax):
        ax.set_ylabel(
            " ".join(
                [
                    _("power"),
                    "({})".format(_("normed")) if self.normed else "",
                    "({})".format(_("detrended")) if self.detrend else "",
                    " ".join(
                        r"[${}^2${}]".format(
                            unit, "/{}".format(_("Hz")) if self.normed else ""
                        )
                        for unit in sorted(
                            u if u is not None else _("unitless")
                            for u in ax.units
                        )
                    ),
                ]
            )
        )

    @LinesAnimator.register_method("matching-axes-found")
    def set_units(self, ax, props={}):
        """
        Update the figure with a single dataset

        Args:
            ax (matplotlib.axes.Axes): the axes
            props (dict, optional): properties of the current dataset
        """
        unit = props.get("unit")
        if not hasattr(ax, "units"):
            ax.units = set()
        if unit not in ax.units:
            ax.units.add(unit)
            self.update_ylabel(ax)

    @LinesAnimator.register_method("axes-added")
    def set_axes_labels(self, ax):
        ax.set_ylabel(_("power"))
        ax.set_xlabel(_("frequency [{}]").format(_("Hz")))

    @LinesAnimator.register_method("after-each-frame")
    def update_data_rate_in_legends(self, *args, **kwargs):
        """
        If :any:`show_data_rate` is ``True``, update the displayed data rate in
        each line's legend.
        """
        now = datetime.datetime.utcnow()
        for ax in self.figure.axes:
            for line in ax.get_lines():
                if self.show_data_rate:
                    x = line.times
                    tdiff = (
                        line.times.max() - line.times.min()
                    ) / np.timedelta64(1, "s")
                    rate = max(x.size - 1, 0) / tdiff if tdiff > 0 else 0
                    rate_val, rate_str = (
                        (1 / rate, _("sec/ds"))
                        if (0 < rate < 1)
                        else (rate, _("Hz"))
                    )
                    rate_text = "({:.1f} {})".format(rate_val, rate_str)
                    rate_text_pattern = r"\([^)]+?({}|{})\)".format(
                        re.escape(_("Hz")), re.escape(_("sec/ds"))
                    )
                    old_label = new_label = line.get_label()
                    if re.search(rate_text_pattern, old_label):
                        new_label = re.sub(
                            rate_text_pattern, rate_text, old_label
                        )
                    else:
                        new_label += " " + rate_text
                    if old_label != new_label:
                        line.set_label(new_label)
                        ax.legend()



# system modules
import io
import logging

# internal modules
import polt
from polt.l10n import _
from polt.config import DEFAULT_CONFIG_FILES
from polt.cli.commands.main import cli

# external modules
import click

logger = logging.getLogger(__name__)


@cli.command(
    help="\n\n".join(
        (
            _("managing configuration"),
            _(
                "The configuration files {files} are read by default "
                "if they exist "
                "unless the top-level --no-config option was specified."
            ).format(files=polt.l10n.join(map(repr, DEFAULT_CONFIG_FILES))),
            _(
                "Without options, this command just dumps the current "
                "configuration."
            ),
        )
    ),
    short_help=_("managing configuration"),
)
@click.option(
    "-r",
    "--read",
    help=_("read a configuration file"),
    metavar=_("file").upper(),
    type=click.File("r"),
)
@click.option(
    "-s",
    "--save",
    help=_("save the current configuration to file"),
    metavar=_("file").upper(),
    type=click.File("w"),
)
@click.pass_context
def config(ctx, read, save):
    if read:
        logger.info(
            _("reading configuration file {file}").format(file=repr(read.name))
        )
        ctx.obj["config"].read_file(read)
    if save:
        logger.info(
            _("saving current configuration to file {file}").format(
                file=repr(save.name)
            )
        )
        ctx.obj["config"].write(save)
    if not (read or save):
        buf = io.StringIO()
        ctx.obj["config"].write(buf)
        click.echo(buf.getvalue().strip())



# system modules
import logging
import re

# internal modules
from polt.filter import Filter
from polt import utils

# external modules
import numpy as np


logger = logging.getLogger(__name__)


class SelectionFilter(Filter):
    """
    Filter for dropping or keeping only specific data
    """

    @property
    def when_quantity(self):
        """
        Only operate on quantities matching this regular expression

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o when-quantity=PATTERN``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_quantity
        except AttributeError:
            self._when_quantity = re.compile(".*")
        return self._when_quantity

    @when_quantity.setter
    def when_quantity(self, new):
        self._when_quantity = re.compile(
            str(new) if new else r".*", flags=re.IGNORECASE
        )

    @property
    def when_unit(self):
        """
        Only operate on units matching this regular expression

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o when-unit=PATTERN``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_unit
        except AttributeError:
            self._when_unit = re.compile(".*")
        return self._when_unit

    @when_unit.setter
    def when_unit(self, new):
        self._when_unit = re.compile(
            str(new) if new else r".*", flags=re.IGNORECASE
        )

    @property
    def when_key(self):
        """
        Only operate on keys matching this regular expression

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o when-key=PATTERN``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_key
        except AttributeError:
            self._when_key = re.compile(".*")
        return self._when_key

    @when_key.setter
    def when_key(self, new):
        self._when_key = re.compile(
            str(new) if new else r".*", flags=re.IGNORECASE
        )

    @property
    def action(self):
        """
        What to do with the matched data. Either ``"drop"`` or ``"keep"``
        (default).
        """
        try:
            return self._action
        except AttributeError:
            self._action = "keep"
        return self._action

    @action.setter
    def action(self, new):
        options = ("keep", "drop")
        if new not in options:
            raise ValueError(f"action must be either {options}, not {new}")
        self._action = new

    def update_dataset(self, dataset):
        """
        ...

        Args:
            dataset (dict): the dataset to modify

        Returns:
            dict or None: the modified dataset

        """
        matched = []
        data = dataset.get("data", {})
        for key, val in data.items():
            keyiter = iter(utils.to_tuple(key))
            props = {
                p: next(keyiter, None) for p in ("quantity", "unit", "key")
            }
            regexes = {
                p: getattr(self, f"when_{p}", re.compile(r".*")) for p in props
            }
            matches = {p: regexes[p].search(v or "") for p, v in props.items()}
            if all(matches.values()):
                matched.append(key)
        for k in (
            (set(data) - set(matched)) if self.action == "keep" else matched
        ):
            data.pop(k, None)
        return dataset



# system modules

# internal modules
import polt.cli.commands.config
import polt.cli.commands.add_source
import polt.cli.commands.add_filter
import polt.cli.commands.live
import polt.cli.commands.generate

# external modules



# system modules
import importlib
import re

# internal modules
import polt
from polt.l10n import _

# external modules
import click


def option_spec(ctx, param, value):
    part = iter(value.split("=", maxsplit=1))
    attr, value = next(part, ""), next(part, True)
    if attr:
        return re.sub(r"\s+", "-", attr), value
    else:
        raise click.BadParameter(
            _(
                "specify an option like {option}={value}, not like {given}"
            ).format(
                option=_("option").upper(),
                value=_("value").upper(),
                given=repr(value),
            ),
            param=param,
            ctx=ctx,
        )


def options_spec(ctx, param, value):
    return dict(map(lambda x: option_spec(ctx, param, x), value))


def python_class_spec(aliases={}):
    def spec(ctx, param, value):
        if value is None:
            return value
        matching_alias_cls = next(
            (cls for al, cls in aliases.items() if value in al), None
        )
        if matching_alias_cls:
            cls = matching_alias_cls
        else:
            parts = value.split(".")
            try:
                package, classname = ".".join(parts[:-1]), parts[-1]
                if package:
                    module = importlib.import_module(package)
                    cls = getattr(module, classname)
                    assert isinstance(cls, type), _(
                        "{obj} is not a type but of type {type}"
                    ).format(obj=cls, type=type(cls))
                else:
                    cls = aliases[classname]
            except (
                IndexError,
                ImportError,
                AttributeError,
                AssertionError,
                ValueError,
                KeyError,
            ) as e:
                raise click.BadParameter(
                    _(
                        "class path specification {spec} "
                        "is neither a native alias (like {native_aliases}) "
                        "nor an importable class specification ({error})"
                        ""
                    ).format(
                        spec=repr(value),
                        native_aliases=polt.l10n.join(
                            polt.extensions.EntryPointExtensions.short_aliases(
                                aliases
                            ),
                            last_sep=" {} ".format(_("or")),
                        ),
                        error=e,
                    ),
                    ctx=ctx,
                    param=param,
                )
        path = ".".join((cls.__module__, cls.__name__))
        return path

    return spec



# system modules

# internal modules
from polt.parser.parser import Parser

# external modules


class MosquittoParser(Parser):
    """
    A simple parser that reads output of the `mosquitto_sub
    <https://mosquitto.org/man/mosquitto_sub-1.html>`_ MQTT cli client invoked
    with the ``-v`` flag.

    Example
    =======

    .. code-block:: sh

        mosquitto_sub -h your.mqtt-broker.org -t '#' -v
        # output like:
        # TOPIC                            VALUE UNIT
        # house/garden/sensor1/temperature 32.1  °C
        # house/cellar/sensor2/humidity    41    %
        # ...

        # feed this into polt like this:
        mosquitto_sub -h your.mqtt-broker.org -t '#' -v \
            | polt add-source -p mosquitto live
    """

    @property
    def data(self):
        for line in self.f:
            topic, message = (
                line.decode(errors="ignore")
                if hasattr(line, "decode")
                else line
            ).split(maxsplit=1)
            message = message.strip()
            topicsplit = topic.split("/")
            key, quantity = "/".join(topicsplit[:-1]), topicsplit[-1]
            try:
                value, unit = message.split(maxsplit=1)
            except (ValueError, TypeError):
                unit = None
                value = message
            try:
                yield {(quantity, unit, key): float(value)}
            except ValueError:
                continue



# system modules
import functools
import collections
import time
import itertools
import logging
import inspect
import math
from abc import abstractmethod

# internal modules
from polt.l10n import _
from polt.filter import Filter
from polt.registry import FunctionRegistry

# external modules
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

logger = logging.getLogger(__name__)


class Animator(FunctionRegistry):
    """
    The :class:`Animator` is an abstract base class for animations. Subclasses
    need to override just the :meth:`Animator.update_figure` method.

    Args:
        figure (matplotlib.figure.Figure, optional): the figure to animate on
        buffer (list-like, optional): the buffer to retrieve
            data from
        filters (sequence of Filter objects, optional): filters to modify
            datasts with
    """

    def __init__(self, figure=None, buffer=None, filters=None):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    def modifies_animation(decorated_function):
        """
        Decorator for methods that require resetting the animation
        """

        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            if hasattr(self, "_animation"):
                logger.debug(
                    _("Clearing animation before {}").format(
                        decorated_function
                    )
                )
                del self._animation
            return decorated_function(self, *args, **kwargs)

        return wrapper

    @property
    def buffer(self):
        """
        The buffer to retrieve data from

        :type: list-like
        """
        try:
            self._buffer
        except AttributeError:  # pragma: no cover
            self._buffer = list()
        return self._buffer

    @buffer.setter
    def buffer(self, new):
        self._buffer = new

    @property
    def filters(self):
        """
        The filters to modify datasets with

        :type: list-like
        """
        try:
            self._filters
        except AttributeError:  # pragma: no cover
            self._filters = list()
        return self._filters

    @filters.setter
    def filters(self, new):
        self._filters = new

    @property
    def interval(self):
        """
        The frame update interval

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o interval=MILLISECONDS``

        :type: :any:`int`
        """
        try:
            self._interval
        except AttributeError:
            self._interval = 200
        return self._interval

    @interval.setter
    def interval(self, new_interval):
        self._interval = max(1, int(new_interval))

    @property
    def max_fps(self):
        """
        The maximum update framerate in frames per second (Hz). Internally this
        sets :any:`interval`.

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o max-fps=FRAMERATE``

        :getter: return the inverse of :attr:`interval`
        :setter: set :attr:`interval` to the inverse of the given value in
            millisecons capped below ``1e-2`` which means 100 seconds per frame
            (very slow...)
        :type: :class:`float`
        """
        return 1 / self.interval

    @max_fps.setter
    def max_fps(self, new):
        self.interval = 1e3 / max(1e-2, float(new))

    @property
    def show_fps(self):
        """
        Whether to show the current plot framerate in the title.
        Default is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -o show-fps=yes|no``

        :type: :class:`bool`
        """
        try:
            self._show_fps
        except AttributeError:
            self._show_fps = False
        return self._show_fps

    @show_fps.setter
    def show_fps(self, new):
        self._show_fps = new == "yes" if isinstance(new, str) else bool(new)

    @property
    def figure(self):
        """
        The figure to animate on

        :type: :class:`matplotlib.figure.Figure`
        """
        try:
            self._figure
        except AttributeError:  # pragma: no cover
            fig = plt.figure()
            fig.canvas.mpl_connect(
                "key_press_event",
                lambda e: self.call_registered_functions(
                    "key-pressed", event=e
                ),
            )
            self._figure = fig
        return self._figure

    @figure.setter
    @modifies_animation
    def figure(self, new):
        self._figure = new

    @property
    def animation(self):
        """
        The underlying animation

        :type: :class:`matplotlib.animation.FuncAnimation`
        """
        try:
            self._animation
        except AttributeError:
            self._animation = FuncAnimation(
                fig=self.figure,
                init_func=self.clear_figure,
                func=self.update_figure,
                frames=self.dataset,
                interval=self.interval,
                repeat=False,
            )
        return self._animation

    @property
    def paused(self):
        """
        Whether the animation is currently paused or not.
        During pause, recieved data is dropped if
        :attr:`Animator.drop_on_pause` is ``True``.

        :type: :class:`bool`
        """
        try:
            self._paused
        except AttributeError:
            self._paused = False
        return self._paused

    @paused.setter
    def paused(self, new):
        new = bool(new)
        if self.paused != new:
            if new:
                self.figure.canvas.set_window_title(
                    "{} ({})".format(_("polt").title(), _("paused"))
                )
            else:
                self.figure.canvas.set_window_title(_("polt").title())
        self._paused = new

    @property
    def drop_on_pause(self):
        """
        Whether to drop recieved data if the animation is paused. The default
        is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -o drop-on-pause=yes|no``

        :type: :class:`bool`
        """
        try:
            self._drop_on_pause
        except AttributeError:
            self._drop_on_pause = False
        return self._drop_on_pause

    @drop_on_pause.setter
    def drop_on_pause(self, new):
        self._drop_on_pause = (
            new == "yes" if isinstance(new, str) else bool(new)
        )

    @property
    def fps_buffer(self):
        """
        Buffer to calculate the fps
        """
        try:
            self._fps_buffer
        except AttributeError:
            self._fps_buffer = collections.deque(maxlen=30)
        return self._fps_buffer

    def pause(self):
        """
        Pause the animation
        """
        self.paused = True

    def resume(self):
        """
        Resume the animation
        """
        self.paused = False

    def toggle_pause(self):
        """
        Change the paused state
        """
        if self.paused:
            logger.info(_("Resuming"))
        else:
            logger.info(_("Pausing"))
            if self.drop_on_pause:
                logger.info(_("Recieved data is dropped form now on."))
            else:
                logger.info(_("Data recording keeps going."))
        self.paused = not self.paused

    @FunctionRegistry.register_method("key-pressed")
    def toggle_pause_key_callback(self, event):
        """
        Callback to pause the animation with :any:`toggle_pause` if the pressed
        key was the spacebar.

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == " ":
            self.toggle_pause()

    @FunctionRegistry.register_method("key-pressed")
    def toggle_show_fps(self, event):
        """
        Callback to toggle :any:`show_fps` if the pressed key was a capital
        ``F``.

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == "F":
            self.show_fps = not self.show_fps
            self.figure.canvas.set_window_title(_("polt").title())

    def clear_figure(self):
        """
        Clear the :attr:`figure` by calling
        :meth:`matplotlib.figure.Figure.clear`.
        """
        self.figure.clear()
        self.figure.canvas.set_window_title(_("polt").title())

    @abstractmethod
    def update_figure(self, buffer):
        """
        Update the :attr:`figure` with a sequence of datasets from the
        :attr:`Animator.buffer`.

        .. note::

            This is an :func:`abc.abstractmethod`, so subclasses have to
            override this.

        Args:
            buffer (sequence): the buffer containing the datasets
                to update the :attr:`figure` with.
        """

    @property
    def dataset(self):
        """
        Endless-loop generator which :class:`list`-ifies the
        :attr:`Animator.buffer`, clears it and yields **another generator**
        yielding non-empty datasets filtered with :any:`filters`.  If
        :any:`paused` and :any:`drop_on_pause` is ``True``, drop the data.

        The following registry hooks are executed:

        frame-before-data-check
            first thing in each iteration

        frame-during-pause
            on each iteration but only during pause before dropping the
            :any:`buffer`

        frame-when-not-paused
            on each iteration but only during pause before reading and clearing
            the :any:`buffer`
        """
        filter_chain = Filter.chain(*self.filters)
        while True:
            self.call_registered_functions("frame-before-data-check")
            if self.paused:
                self.call_registered_functions("frame-during-pause")
                if self.drop_on_pause:
                    # empty the buffer
                    del self.buffer[:]
            else:
                self.call_registered_functions("frame-when-not-paused")
                # read the whole buffer
                buf = list(self.buffer)
                # empty the buffer
                del self.buffer[:]
            # remember the time we yielded this dataset
            self.fps_buffer.append(time.time())
            # modify datasets with the filters and ignore empty datasets
            yield filter(bool, map(filter_chain, buf))

    def run(self):
        """
        Run the animation by calling :func:`matplotlib.pyplot.show`
        """
        self.animation
        logger.debug(_("Showing the plot"))
        plt.show()
        logger.debug(_("Plot was closed"))

    @FunctionRegistry.register_method("frame-before-data-check")
    def update_window_title(self):
        if self.show_fps:
            n = len(self.fps_buffer)
            if n > 1:
                fps = n / (max(self.fps_buffer) - min(self.fps_buffer))
                self.figure.canvas.set_window_title(
                    " ".join(
                        (
                            _("polt").title(),
                            "({:.0f} {})".format(fps, _("fps")),
                        )
                    )
                )



# system modules
import functools
import datetime
import logging
import inspect
import operator
import sys

# internal modules
from polt.l10n import _
from polt.animator.lines import LinesAnimator
from polt.utils import *

# external modules
import numpy as np
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)


class HistAnimator(LinesAnimator):
    """
    Animator to display recieved data as lines.

    Args:
        args, kwargs: further arguments handed to the
            :any:`LinesAnimator` constructor.
        bins (int, optional): the number of bins to use.
        normed (bool, optional): whether to display density instead of
            frequency. Default is ``False``.
        time_frame (int, optional): displayed time frame in seconds. By default
            (``None``), all times are diplayed.
    """

    def __init__(
        self, *args, bins=None, normed=None, time_frame=None, **kwargs
    ):
        LinesAnimator.__init__(self, *args, **kwargs)
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @property
    def bins(self):
        """
        How many bins to use. Default is 10.

        .. note::

            Set this option from the command-line via
            ``polt live -a hist -o bins=NUMBER``

        :type: :class:`str` or ``None``
        """
        try:
            self._bins
        except AttributeError:
            self._bins = 10
        return self._bins

    @bins.setter
    def bins(self, new):
        self._bins = max(2, abs(int(new)))

    @property
    def normed(self):
        """
        Whether to norm the histogram. By default (``False``), the
        absolute frequency is used.

        .. note::

            Set this option from the command-line via
            ``polt live -a hist -o normed=yes|no``

        :type: :class:`str` or ``None``
        """
        try:
            self._normed
        except AttributeError:
            self._normed = False
        return self._normed

    @normed.setter
    def normed(self, new):
        self._normed = new == "yes" if isinstance(new, str) else bool(new)

    @property
    def time_frame(self):
        """
        Time frame to use in seconds. Data outside that window
        will be dropped.

        .. note::

            Set this option from the command-line via
            ``polt live -a hist -o time-frame=SECONDS``

        :type: :class:`int` or ``None``
        """
        try:
            self._time_frame
        except AttributeError:
            self._time_frame = None
        return self._time_frame

    @time_frame.setter
    def time_frame(self, new):
        self._time_frame = max(sys.float_info.min, int(new))

    @property
    def show_data_rate(self):
        """
        Whether to show the displayed data rate in the legend label. The
        default is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -a hist -o show-data-rate=yes|no``

        :type: :class:`bool`
        """
        try:
            self._show_data_rate
        except AttributeError:
            self._show_data_rate = False
        return self._show_data_rate

    @show_data_rate.setter
    def show_data_rate(self, new):
        self._show_data_rate = (
            new == "yes" if isinstance(new, str) else bool(new)
        )

    def histogram(self, *args, **kwargs):
        """
        Call :any:`numpy.histogram` with default arguments.

        Args:
            **kwargs: overwriting keyword arguments

        Returns:
            sequence : the same as :any:`numpy.histogram`
        """
        hist_kwargs = {"bins": self.bins, "density": self.normed}
        hist_kwargs.update(kwargs)
        return np.histogram(*args, **hist_kwargs)

    def new_line_coords(self, values, dataset={}):
        time_recieved_utc = np.datetime64(
            dataset.get("time_recieved_utc", datetime.datetime.utcnow())
        )
        values = np.array(values)
        times = np.repeat(np.datetime64(time_recieved_utc), values.size)
        freq, bins = self.histogram(values)
        x = np.repeat(bins, 2)
        y = np.concatenate([np.array([0]), np.repeat(freq, 2), np.array([0])])
        return x, y, {}, {"times": times, "values": values}

    def update_line(self, line, values, dataset={}):
        time_recieved_utc = np.datetime64(
            dataset.get("time_recieved_utc", datetime.datetime.utcnow())
        )
        values = np.array(values)
        times = np.repeat(np.datetime64(time_recieved_utc), values.size)
        line.values = np.append(line.values, values)
        line.times = np.append(line.times, times)
        freq, bins = self.histogram(line.values)
        line.set_xdata(np.repeat(bins, 2))
        line.set_ydata(
            np.concatenate([np.array([0]), np.repeat(freq, 2), np.array([0])])
        )

    @LinesAnimator.register_method("key-pressed")
    def toggle_normed(self, event):
        """
        Callback to toggle :any:`normed` if the pressed
        key was an ``n``.

        .. hint::

            This means, you can press ``n`` when using the :any:`HistAnimator`
            to switch between normed and absolute histograms.

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == "n":
            self.normed = not self.normed
            for ax in self.figure.axes:
                self.set_axes_labels(ax)

    @LinesAnimator.register_method("before-line-update")
    def drop_old_values(self, line):
        if self.time_frame:
            outside = (
                np.datetime64(datetime.datetime.utcnow()) - line.times
            ) / np.timedelta64(1, "s") > self.time_frame
            if outside.any():
                outside_ind = np.where(outside)[0]
                line.values = np.delete(line.values, outside_ind)
                line.times = np.delete(line.times, outside_ind)

    @LinesAnimator.register_method("matching-axes-found")
    def set_units(self, ax, props={}):
        """
        Update the figure with a single dataset

        Args:
            ax (matplotlib.axes.Axes): the axes
            props (dict, optional): properties of the current dataset
        """
        unit = props.get("unit")
        if not hasattr(ax, "units"):
            ax.units = set()
        if unit not in ax.units:
            ax.units.add(unit)

    @LinesAnimator.register_method("axes-added")
    def set_axes_labels(self, ax):
        ax.set_ylabel(_("density") if self.normed else _("frequency"))
        ax.set_xlabel(_("value"))

    @LinesAnimator.register_method("after-each-frame")
    def adjust_view(self, *args, **kwargs):
        """
        Adjust the view with :any:`LinesAnimator.fit_view` for both axes.
        """
        self.fit_view(axes="xy")

    @LinesAnimator.register_method("after-each-frame")
    def update_data_rate_in_legends(self, *args, **kwargs):
        """
        If :any:`show_data_rate` is ``True``, update the displayed data rate in
        each line's legend.
        """
        now = datetime.datetime.utcnow()
        for ax in self.figure.axes:
            for line in ax.get_lines():
                if self.show_data_rate:
                    x = line.get_xdata()
                    if (
                        line.last_extrapolated
                        if hasattr(line, "last_extrapolated")
                        else False
                    ):
                        x = x[:-1]
                    tdiff = (max(x.max(), now) - x.min()).total_seconds()
                    rate = max(x.size - 1, 0) / tdiff if tdiff > 0 else 0
                    rate_val, rate_str = (
                        (1 / rate, _("sec/ds"))
                        if (0 < rate < 1)
                        else (rate, _("Hz"))
                    )
                    rate_text = "({:.1f} {})".format(rate_val, rate_str)
                    rate_text_pattern = r"\([^)]+?({}|{})\)".format(
                        re.escape(_("Hz")), re.escape(_("sec/ds"))
                    )
                    old_label = new_label = line.get_label()
                    if re.search(rate_text_pattern, old_label):
                        new_label = re.sub(
                            rate_text_pattern, rate_text, old_label
                        )
                    else:
                        new_label += " " + rate_text
                    if old_label != new_label:
                        line.set_label(new_label)
                        ax.legend()



# system modules
import inspect
import logging

from abc import ABC, abstractproperty

# internal modules
from polt.l10n import _

# external modules

logger = logging.getLogger(__name__)


class Parser(ABC):
    """
    Abstract base class for parsers.

    Args:
        f (file-like object, optional): the connection to read from
        name (str, optional): human-readable description of the parser
    """

    def __init__(self, f=None, name=None):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @property
    def name(self):
        """
        Human-readable description of the parser

        :type: :class:`str`
        :getter: If no :attr:`name` was set yet, attempt to return a
            meaningful name based on :attr:`f`.
        """
        try:
            return self._name
        except AttributeError:
            return _("{} on {}").format(
                type(self).__name__,
                self.f.name if hasattr(self.f, "name") else repr(self.f),
            )

    @name.setter
    def name(self, new):
        self._name = str(new)

    @property
    def f(self):
        """
        The connection to read from
        """
        try:
            self._f
        except AttributeError:  # pragma: no cover
            self._f = None
        return self._f

    @f.setter
    def f(self, new):
        self._f = new

    @abstractproperty
    def data(self):
        """
        Generator yielding the next dataset. This is an
        :meth:`abc.abstractproperty`, thus subclasses have to override this.

        Yields:
            dict : dictionary like

                .. code:: python

                    {
                        quantity_spec1: value1,
                        quantity_spec2: value2,
                        quantity_spec3: value3,
                        ...
                    }

                quantity_spec
                    either a :class:`str` containing the name of the quantity
                    or a :class:`tuple` ``(name, unit, key)`` specifying the
                    quantity's name, unit and key
                value
                    either a single value or a sequence of values
        """
        pass



# system modules
import collections
import functools
import textwrap
from abc import ABCMeta

# internal modules

# external modules


class RegisteringType(ABCMeta):
    """
    Metaclass adding callables that are marked as ”registered” (i.e. have a
    ``_registered_slot`` attribute) to a ``registry`` class attribute. This
    :attr:`RegisteringType.registry` class attribute is a mapping of classes to
    mappings of slot definitions to functions.

    This metaclass is used for the :any:`FunctionRegistry` class.

    .. note::

        This class inherits from :any:`abc.ABCMeta` so all classes using this
        class as metaclass will behave like :any:`abc.ABC`.

    .. note::

        Implementation based on `this StackOverflow answer
        <https://stackoverflow.com/a/3054505>`_.

    """

    registry = collections.defaultdict(lambda: collections.defaultdict(list))
    """
    Mapping of classes to mappings of slot definitions to functions:

    .. code-block:: python

        registry[class][slot] = [callable1, callable2, ...]

    :type: :class:`collections.defaultdict`
    """

    def __init__(cls, name, bases, attrs):
        for key, val in attrs.items():
            if hasattr(val, "__call__") and hasattr(val, "_registered_slot"):
                cls.registry[cls][val._registered_slot].append(val)


class FunctionRegistry(metaclass=RegisteringType):
    """
    Class with metaclass :any:`RegisteringType` that implements decorators
    to register functions in **class-based** (**not** per instance) slots which
    can be called with :any:`call_registered_functions`.

    .. note::

        Implementation based on `this StackOverflow answer
        <https://stackoverflow.com/a/3054505>`_.

    """

    @classmethod
    def registering_decorator(
        cls, _slot, *args, _on=None, _doc=True, **kwargs
    ):
        """
        Create a registering decorator

        Args:
            _slot (str): the slot to register the function in
            _doc (bool, optional): whether the decorator will prepend a note on
                how the decorated function was registered to its docstring.
                Default is yes.
            _on (str or None, optional): How the decorator will register the
                function:

                ``None``
                    put the decorated function directly into
                    :any:`RegisteringType.registry`
                ``"instance"``
                    register the decorated function so that it will be called
                    with the instance as first argument
                ``"class"``
                    register the decorated function so that it will be called
                    with the class as first argument

            *args: further positional arguments for the function
            **kwargs: further keyword arguments for the function
        """

        def decorator(f):
            assert hasattr(
                f, "__call__"
            ), "This decorator can only be used to decorate callable objects."
            f._registered_slot = _slot
            f._registered_args = args
            f._registered_kwargs = kwargs
            if _doc:
                f.__doc__ = "**Note**: {note_text}\n\n{old_doc}".format(
                    note_text=textwrap.indent(
                        (
                            "This function is registered for the ``{slot}`` "
                            "slot of the :any:`{cls}` class."
                        ).format(
                            slot=_slot,
                            cls=".".join((cls.__module__, cls.__name__)),
                        ),
                        " " * 4,
                    ),
                    old_doc=f.__doc__ or "",
                )
            if _on is None:
                cls.registry[cls][_slot].append(f)
            else:
                f._registered_on = _on
            return f

        return decorator

    @classmethod
    def register_function(cls, _slot, *args, **kwargs):
        """
        Return a decorator registering an arbitrary function to a given slot

        Args:
            _slot (str): the slot to register the function in
            *args: further positional arguments for the function
            **kwargs: further keyword arguments for the function
        """
        return cls.registering_decorator(_slot, *args, **kwargs)

    @classmethod
    def register_method(cls, _slot, *args, **kwargs):
        """
        Return a decorator registering an instance method to a given slot.

        Args:
            _slot (str): the slot to register the method in
            *args: further positional arguments for the function
            **kwargs: further keyword arguments for the function
        """
        return cls.registering_decorator(
            _slot, *args, _on="instance", **kwargs
        )

    @classmethod
    def register_classmethod(cls, _slot, *args, **kwargs):
        """
        Return a decorator registering a classmethod to a given slot.

        Args:
            _slot (str): the slot to register the method in
            *args: further positional arguments for the function
            **kwargs: further keyword arguments for the function
        """
        return cls.registering_decorator(_slot, *args, _on="class", **kwargs)

    def call_registered_functions(self, slot, *args, **kwargs):
        """
        Call all functions of a given slot with their configured arguments.
        Functions registered from classes which this class does not inherit
        from are not executed.

        Args:
            slot (str): the slot to execute the registered functions from
            args, kwargs: further arguments to pass to the functions

        Returns:
            dict : mapping of called functions to their return values
        """
        ret = {}
        for cls, slots in type(self).registry.items():
            if isinstance(self, cls):
                for func in slots.get(slot, tuple()):
                    first_arg = {
                        "instance": (self,),
                        "class": (type(self),),
                    }.get(getattr(func, "_registered_on", None), tuple())
                    reg_args = getattr(func, "_registered_args", tuple())
                    reg_kwargs = getattr(func, "_registered_kwargs", {})
                    func_kwargs = reg_kwargs.copy()
                    func_kwargs.update(kwargs)
                    ret[func] = func(
                        *(first_arg + reg_args + args), **func_kwargs
                    )
        return ret



# system modules
import pkg_resources
import warnings
import logging

# internal modules
from polt.version import *
import polt.l10n
from polt.l10n import _
import polt.animator
import polt.streamer
import polt.parser
import polt.utils
import polt.config

# external modules

logger = logging.getLogger(__name__)


class EntryPointExtensions:
    """
    Class for detection of entry point extensions for :mod:`polt`
    """

    entry_point_subclasses = {
        POLT_PARSER_ENTRY_POINT: polt.parser.parser.Parser,
        POLT_ANIMATOR_ENTRY_POINT: polt.animator.Animator,
    }

    @classmethod
    def get(cls, name, subcls=None, aliases=False):
        """
        Check entry points. Invalid entry points are skipped with a warning.
        Entry points pointing to classes not inheriting from a given class are
        still included but with a warning.

        Args:
            name (str): the name of the entry point
            cls (class, optional): class to check inheritance for. Defaults to
                :class:`object`.
            aliases (bool, optional): Whether to pass the result through
                :meth:`aliases_from_entry_points` to determine the aliases.
                Default is ``False``.

        Returns:
            dict : mapping of :class:`pkg_resources.EntryPoint` s (if
            ``aliases`` is ``False``, tuples of unique alias strings otherwise)
            to the respective classes
        """
        if subcls is None:
            subcls = cls.entry_point_subclasses.get(name, object)
        entry_points = {}
        for entry_point in pkg_resources.iter_entry_points(name):
            try:
                entry_point_cls = entry_point.load()
            except BaseException as e:
                warnings.warn(
                    _(
                        "Could not load {entry_point} entry point {name} "
                        "from {dist}: {err}"
                    ).format(
                        entry_point=repr(name),
                        name=repr(entry_point.name),
                        err=e,
                        dist=repr(
                            entry_point.dist or _("unknown distribution")
                        ),
                    )
                )
                continue
            try:
                if not issubclass(entry_point_cls, subcls):
                    warnings.warn(
                        _(
                            "{entry_point} entry point named {name} "
                            "from {dist} points to class "
                            "{entry_point_cls} which is not a "
                            "{subcls} subclass "
                            "and thus might not work"
                        ).format(
                            entry_point=repr(name),
                            name=repr(entry_point.name),
                            entry_point_cls=entry_point_cls,
                            subcls=subcls,
                            dist=repr(
                                entry_point.dist or _("unknown distribution")
                            ),
                        )
                    )
            except BaseException as e:
                warnings.warn(
                    _(
                        "Skipping {entry_point} entry point named {name} "
                        "from {dist} pointing to {obj} which is "
                        "obviously not a type: {err}"
                    ).format(
                        entry_point=repr(subcls),
                        name=repr(entry_point.name),
                        obj=repr(entry_point_cls),
                        err=e,
                        dist=repr(
                            entry_point.dist or _("unknown distribution")
                        ),
                    )
                )
                continue
            matching_ep = next(
                filter(lambda ep: ep.name == entry_point.name, entry_points),
                None,
            )
            if matching_ep:
                warnings.warn(
                    _(
                        "{dist1} and {dist2} both define a {entry_point} "
                        "entry point {name}."
                    ).format(
                        dist1=repr(
                            matching_ep.dist or _("unknown distribution")
                        ),
                        dist2=repr(
                            entry_point.dist or _("unknown distribution")
                        ),
                        entry_point=repr(name),
                        name=repr(entry_point.name),
                    )
                )
            entry_points[entry_point] = entry_point_cls
        return (
            cls.aliases_from_entry_points(entry_points)
            if aliases
            else entry_points
        )

    @staticmethod
    def aliases_from_entry_points(entry_points):
        """
        Create mapping of aliases to classes from

        Args:
            entry_points (dict): mapping of entry points to the respective
                classes

        Returns:
            dict : mapping of tuples of unique alias strings to classes
        """
        final_aliases = {}
        for entry_point, final_cls in sorted(
            entry_points.items(),
            key=lambda x: (
                "0" if x[0].module_name.startswith("polt") else str(x[0])
            ),
        ):

            aliases = []
            full = (
                ".".join(filter(bool, (entry_point.module_name, a)))
                for a in entry_point.attrs
            )
            for alias in filter(
                bool,
                (entry_point.name,) + tuple(entry_point.attrs) + tuple(full),
            ):
                if any(alias in a for a in final_aliases):
                    continue
                aliases.append(alias)
            final_aliases[tuple(aliases)] = final_cls
        return final_aliases

    @staticmethod
    def short_aliases(aliases):
        """
        Generate a list of aliases

        Args:
            aliases (sequence of sequences): sequence of sequences with alias
                strings

        Returns:
            tuple : the first aliases
        """
        return tuple(
            map(
                repr,
                sorted(filter(bool, (next(iter(a), None) for a in aliases))),
            )
        )



# system modules
import shlex
import collections
import functools

# internal modules

# external modules


def flatten(S):
    """
    Function to recursively flatten a list

    Args:
        S (list): the list of nested lists to flatten

    Returns:
        list : the flattened list
    """
    if S == []:
        return S
    if isinstance(S[0], list):
        return flatten(S[0]) + flatten(S[1:])
    return S[:1] + flatten(S[1:])


def normalize_cmd(cmd):
    """
    Normalise a command

    Args:
        cmd (str): the command to normalize

    Returns:
        str : the normalized command
    """
    return " ".join(map(shlex.quote, shlex.split(cmd)))


def str2num(s):
    """
    Convert a string to a number

    Args:
        s (str): the string to parse

    Returns:
        int: if input looked like an integer
        float: if input looked like a floating-point number
        str: the input itself if nothing else worked
    """
    try:
        i = int(s)
    except ValueError:
        i = None
    try:
        f = float(s)
    except ValueError:
        f = None
    return s if i == f is None else (i if i == f else f)


def to_float(o, default=None, default_on_failure=True):
    """
    Convert an object to float

    Args:
        o (object): the object to convert
        default (object, optional): the value to return on failure if
            ``default_on_failure=True``
        default_on_failure (bool, optional): whether to return the value of
            ``default`` on failure. Defaults to ``True``.

    Returns:
        float or ``type(default)``: the converted float. On failure the value
            of ``default`` if ``default_on_failure=True`` or else the original
            value.
    """
    try:
        return float(o)
    except BaseException:
        return default if default_on_failure else o


def to_tuple(o):
    """
    Convert an object to a :class:`tuple`. If not iterable or :class:`str`,
    return a :class:`tuple` with just the object as element.

    Args:
        o (object): the object to convert

    Returns:
        sequence: tuple
    """
    if isinstance(o, str):
        return (o,)
    try:
        return tuple(o)
    except TypeError:
        return (o,)



# system modules
import logging
import re
import itertools
import importlib
import shlex
import shutil
import sys

# internal modules
import polt
from polt.l10n import _
from polt.cli.utils import options_spec, python_class_spec
from polt.cli.commands.main import cli

# external modules
import click

logger = logging.getLogger(__name__)

parser_aliases = polt.extensions.EntryPointExtensions.get(
    polt.version.POLT_PARSER_ENTRY_POINT, aliases=True
)
short_parser_aliases = polt.extensions.EntryPointExtensions.short_aliases(
    parser_aliases
)


def cmd_spec(ctx, param, value):
    if value in ("-", "stdin", None):
        return value
    executable = next(iter(shlex.split(value)), None)
    if shutil.which(executable or ""):
        return value
    else:
        raise click.BadParameter(
            _(
                "could not find executable {executable} "
                "from given command {value}. "
                "If you can execute {executable} from your command-line, "
                "try using the output of `which {executable}` as executable."
            ).format(executable=repr(executable), value=repr(value)),
            param=param,
            ctx=ctx,
        )


@cli.command(help=_("add a new data source"))
@click.option(
    "-p",
    "--parser",
    metavar=_("class").upper(),
    callback=python_class_spec(parser_aliases),
    help=_(
        "parser class to use. "
        "This can be a built-in parser ({native_parsers}) "
        "or a class specification to a subclass of Parser "
        "(like 'mypackage.MyParser')"
    ).format(
        native_parsers=polt.l10n.join(
            short_parser_aliases, last_sep=" {} ".format(_("or"))
        )
    ),
)
@click.option(
    "-c",
    "--cmd",
    metavar=_("command").upper(),
    callback=cmd_spec,
    help=_(
        "the command to use as source for the parser. "
        "The default is to use data from stdin."
    ),
)
@click.option(
    "-e",
    "--encoding",
    metavar=_("encoding").upper(),
    help=_(
        "the encoding to use to decode the data. "
        "The default is no encoding which means to pass raw bytes to "
        "the parser."
    ),
)
@click.option(
    "-o",
    "--option",
    "parser_options",
    metavar="{option}={value}".format(
        option=_("option"), value=_("value")
    ).upper(),
    multiple=True,
    callback=options_spec,
    help=_(
        "further options for the parser. "
        "Multiple specifications of this option are possible. "
        "The parser's attribute (previously snake-cased) {option} "
        "will be set to the string {value}. If {value} is not given, {option} "
        "is set to {value_default}."
    ).format(
        option=_("option").upper(),
        value=_("value").upper(),
        value_default=repr(polt.utils.to_float(True)),
    ),
)
@click.option(
    "--max-rate",
    metavar=_("Hertz").upper(),
    help=_(
        "Maximum data rate (datasets per second in Hertz) "
        "to append to the shared buffer. Default is unlimited. "
        "Use this to if the input stream is too fast to plot."
        "Note that the effective data rate will be lower than this."
    ),
    type=click.FloatRange(min=sys.float_info.min),
)
@click.option(
    "--flush-after",
    metavar=_("seconds"),
    help=_(
        "How many seconds to buffer recieved datasets in the streamers "
        "before appending them to the shared buffer. The default is 0 which "
        "means to flush recieved data right away. Increasing this value "
        "can help if the input data arrives too fast for the plot to catch up."
    ),
    type=click.FloatRange(min=0, clamp=True),
)
@click.pass_context
def add_source(
    ctx, parser, cmd, parser_options, encoding, max_rate, flush_after
):
    if not (parser or cmd):
        raise click.UsageError(
            _("Specify at least either --parser or --cmd"), ctx=ctx
        )
    if parser is None:
        parser = polt.parser.numberparser.NumberParser
        parser = ".".join((parser.__module__, parser.__name__))
    if cmd is None:
        cmd = "-"
    config = ctx.obj["config"]
    matching_sections = list(config.matching_source_section(command=cmd))
    for section in matching_sections:
        logger.warning(
            _(
                "Removing previously existing configuration section {section} "
                "because it matches the given command-line source command "
                "{clicmd}."
            ).format(
                clicmd=repr(cmd),
                section=_("{section} (parsing {cmd} with {parser})").format(
                    section=repr(section.name),
                    parser=section.get("parser"),
                    cmd=repr(section.get("command")),
                ),
            )
        )
        config.remove_section(section.name)
    # determine a nice section name
    section_name = "{}:{}".format(
        config.SOURCE_SECTION_PREFIX,
        parser_options.get(
            "name", "stdin" if cmd == "-" else polt.utils.normalize_cmd(cmd)
        ),
    )
    config.add_section(section_name)
    config.update_option(section_name, "command", cmd)
    config.update_option(section_name, "parser", parser)
    config.update_option(section_name, "encoding", encoding)
    for option, value in parser_options.items():
        config[section_name]["parser.{}".format(option)] = value
    streamer_options = {"max-rate": max_rate, "flush-after": flush_after}
    for option, value in streamer_options.items():
        if value is not None:
            config[section_name]["streamer.{}".format(option)] = str(value)



# system modules
import logging
import re
import itertools
import importlib
import shlex
import shutil
import sys

# internal modules
import polt
from polt.l10n import _
from polt.cli.utils import options_spec, python_class_spec
from polt.cli.commands.main import cli

# external modules
import click

logger = logging.getLogger(__name__)

filter_aliases = polt.extensions.EntryPointExtensions.get(
    polt.version.POLT_FILTER_ENTRY_POINT, aliases=True
)
short_filter_aliases = polt.extensions.EntryPointExtensions.short_aliases(
    filter_aliases
)


@cli.command(help=_("add a new data filter"))
@click.option(
    "-f",
    "--filter",
    "filter_spec",
    metavar=_("class").upper(),
    callback=python_class_spec(filter_aliases),
    required=True,
    help=_(
        "filter class to use. "
        "This can be a built-in filter ({native_filters}) "
        "or a class specification to a subclass of Filter "
        "(like 'mypackage.MyFilter')"
    ).format(
        native_filters=polt.l10n.join(
            short_filter_aliases, last_sep=" {} ".format(_("or"))
        )
    ),
)
@click.option(
    "-o",
    "--option",
    "filter_options",
    metavar="{option}={value}".format(
        option=_("option"), value=_("value")
    ).upper(),
    multiple=True,
    callback=options_spec,
    help=_(
        "further options for the filter. "
        "Multiple specifications of this option are possible. "
        "The filter's attribute (previously snake-cased) {option} "
        "will be set to the string {value}. If {value} is not given, {option} "
        "is set to {value_default}."
    ).format(
        option=_("option").upper(),
        value=_("value").upper(),
        value_default=repr(polt.utils.to_float(True)),
    ),
)
@click.pass_context
def add_filter(ctx, filter_spec, filter_options):
    config = ctx.obj["config"]
    section_name = config.add_section(
        "{}:{}".format(config.FILTER_SECTION_PREFIX, filter_spec)
    )
    config.update_option(section_name, "filter", filter_spec)
    for option, value in filter_options.items():
        config[section_name]["filter.{}".format(option)] = value



# system modules
import functools
import collections
import itertools
import logging
import inspect
import datetime
import operator
import re
from abc import abstractmethod

# internal modules
from polt.l10n import _
from polt.animator.subplots import SubPlotsAnimator
from polt.utils import *

# external modules
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.dates import date2num, num2date

logger = logging.getLogger(__name__)


class LinesAnimator(SubPlotsAnimator):
    """
    Animator to display recieved data as lines.

    Args:
        args, kwargs: further arguments handed to the
            :any:`SubPlotsAnimator` constructor.
    """

    @property
    def only_expand_view(self):
        """
        Whether to always expand the view. Default is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -o only-expand-view=yes|no``

        :type: :class:`bool`
        """
        try:
            self._only_expand_view
        except AttributeError:
            self._only_expand_view = False
        return self._only_expand_view

    @only_expand_view.setter
    def only_expand_view(self, new):
        self._only_expand_view = (
            new == "yes" if isinstance(new, str) else bool(new)
        )

    def matching_line(self, props={}, ax=None):
        """
        Generator yielding the next line matching a data properties

        Args:
            props (dict, optional): dictionary of data properties
                with keys

                source
                    name of the data source
                quantity
                    name of the data quantity
                unit
                    data unit
                key
                    data key

            ax(matplotlib.axes.Subplot, optional): the axes to look on.
                Defaults to all axes on :attr:`figure`.

        Yields:
            matplotlib.lines.Lines2D : the next matching line instance
        """
        for cur_ax in self.matching_axes(props) if ax is None else [ax]:
            for line in cur_ax.get_lines():
                if all(
                    props.get(k) == v
                    for k, v in (
                        line.restrictions
                        if hasattr(line, "restrictions")
                        else {}
                    ).items()
                ):
                    yield line

    @abstractmethod
    def new_line_coords(self, values, dataset={}):
        """
        Given values, determine coordinates for a new line.

        .. note::

            This is an :func:`abc.abstractmethod`, so subclasses have to
            override this.

        Args:
            values (sequence of float): the values to use
            dataset (dict, optional): the whole dataset where the values were
                extracted from. This is a dataset from :any:`Animator.buffer`
                as appended by :any:`Streamer.run`. The default is an empty
                dictionary.

        Returns:
            x, y, kwargs, attrs: ``x`` and ``y`` are coordinates as something
            that can be used for :any:`matplotlib.lines.Line2D.set_xdata` and
            :any:`matplotlib.lines.Line2D.set_ydata`. ``kwargs`` is a
            dictionary with further keyword arguments to
            :any:`matplotlib.axes.Axes.plot`. ``attrs`` is a dictionary of
            attributes and values which should be set on the newly created
            line.
        """

    @abstractmethod
    def update_line(self, line, values, dataset={}):
        """
        Update a given line with values

        .. note::

            This is an :func:`abc.abstractmethod`, so subclasses have to
            override this.

        Args:
            line (matplotlib.lines.Line2D): the line to update
            values (sequence of float): the values to use for updating
            dataset (dict, optional): the whole dataset where the values were
                extracted from. This is a dataset from :any:`Animator.buffer`
                as appended by :any:`Streamer.run`. The default is an empty
                dictionary.
        """

    def update_figure_with_dataset(self, dataset):
        """
        Update the figure with a single dataset.

        The following registry hooks are executed:

        matching-axes-found
            When the axes for the line was determined this hook is executed and
            handed the axes as ``ax`` keyword argument.

        line-created
            When a new line was created this hook is executed and handed the
            the line as ``line`` keyword argument.

        before-line-update
            Before a line is updated with new coordinates this hook is executed
            and handed the line as ``line`` keyword argument.

        line-updated
            When a line was updated with new coordinates this hook is executed
            and handed the line as ``line`` keyword argument.

        Args:
            dataset (dict): the dataset to update the :attr:`figure` with
        """
        logger.debug(_("Updating the figure with dataset {}").format(dataset))
        now = datetime.datetime.utcnow()
        parser = dataset.get("parser", None)
        data = dataset.get("data", {})
        for quant, val in data.items():
            if val is None:
                continue
            quantiter = iter((quant,) if isinstance(quant, str) else quant)
            quantity, unit, key = (next(quantiter, None) for i in range(3))
            props = {
                "parser": parser,
                "quantity": quantity,
                "unit": unit,
                "key": key,
            }
            # make sure all values are numeric
            values = tuple(
                filter(
                    lambda x: isinstance(x, float),
                    map(to_float, to_tuple(val)),
                )
            )
            if not values:
                continue
            # determine matching axes
            ax = next(self.matching_axes(props), None)
            if ax is None:
                logger.debug(_("Didn't find matching axes. Adding new."))
                restrictions = {}
                if self.subplots_for is not None:
                    restrictions[self.subplots_for] = props[self.subplots_for]
                ax = self.add_axes(restrictions=restrictions)
            self.call_registered_functions(
                "matching-axes-found", ax=ax, props=props
            )
            # determine matching line
            line = next(self.matching_line(props), None)
            if line:
                self.call_registered_functions("before-line-update", line=line)
                self.update_line(line=line, dataset=dataset, values=values)
                self.call_registered_functions("line-updated", line=line)
            else:
                logger.debug(
                    _("No line on {} matches {}. Plotting new.").format(
                        ax, props
                    )
                )
                coords = iter(
                    self.new_line_coords(values=values, dataset=dataset)
                )
                x, y = next(coords), next(coords)
                extra_plot_kwargs = next(coords, {})
                attrs = next(coords, {})
                plot_kwargs = {
                    "label": " ".join(
                        ([quantity] if quantity else [])
                        + ([key] if key else [])
                        + (["[{}]".format(unit)] if unit else [])
                        + (["({})".format(parser)] if parser else [])
                    )
                }
                plot_kwargs.update(extra_plot_kwargs)
                lines = ax.plot(x, y, **plot_kwargs)
                for cur_line in lines:
                    cur_line.restrictions = {
                        k: v
                        for k, v in props.items()
                        if k not in ax.restrictions
                    }
                    for attr, v in attrs.items():
                        setattr(cur_line, attr, v)
                    self.call_registered_functions(
                        "line-created", line=cur_line
                    )
                ax.legend()

    @SubPlotsAnimator.register_method("before-each-frame")
    def optimal_subplots_layout(self, *args, **kwargs):
        current_ratio = functools.reduce(
            operator.truediv, self.figure.get_size_inches()
        )
        if (
            self.figure.last_ratio
            if hasattr(self.figure, "last_ratio")
            else None
        ) != current_ratio and len(self.figure.axes):
            self.reorder_subplots()
            self.figure.last_ratio = current_ratio

    @SubPlotsAnimator.register_method("key-pressed")
    def toggle_normed(self, event):
        """
        Callback to toggle :any:`only_expand_view` if the pressed
        key was an ``e``.

        .. hint::

            This means, you can press ``e`` when using a :any:`LinesAnimator`
            to toggle axis expanding.

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == "e":
            if self.only_expand_view:
                logger.info(_("Stop always expanding the view"))
            else:
                logger.info(_("Will not only expand the view"))
            self.only_expand_view = not self.only_expand_view

    def update_figure(self, buffer):
        """
        If :any:`paused`, return. Otherwise, :any:`update_figure_with_dataset`
        for all datasets in the given buffer.

        The following registry hooks are executed:

        before-each-frame
            first thing that is done in this function (and before pausing)

        after-each-frame
            last thing that is done in this function

        Args:
            buffer (list-like): the buffer containing the datasets
                to update the :attr:`figure` with.
        """
        # call functions registered for before the frame
        self.call_registered_functions("before-each-frame", buffer)
        if self.paused:
            return
        # update the figure with all buffered datasets
        for dataset in buffer:
            self.update_figure_with_dataset(dataset)
        # call functions registered for after the frame
        self.call_registered_functions("after-each-frame")

    def fit_view(self, axes, only_expand=None):
        """
        Fit the view to the data. If ``only_expand_view`` is ``True``, both
        axis limits are only expanded and never shrinked except when resetting
        the previous respective extremes by changing the axis scale (e.g. by
        hitting the ``l`` or ``k`` key).

        Args:
            axes (sequence of str): iterable of values ``"x"`` and ``"y"``
            only_expand (bool, optional): Whether to only expand
        """
        only_expand = (
            self.only_expand_view if only_expand is None else only_expand
        )
        lim = {
            "x": {"min": "left", "max": "right"},
            "y": {"min": "bottom", "max": "top"},
        }
        for ax in self.figure.axes:
            ax.relim()
            if only_expand:
                for axis in axes:
                    set_lim = getattr(ax, "set_{}lim".format(axis))
                    set_lim_kwargs = {}
                    last_scale_attr = "last_{}scale".format(axis)
                    scale = getattr(ax, "get_{}scale".format(axis))()
                    last_scale = getattr(ax, last_scale_attr, None)
                    for func in (min, max):
                        side = func.__name__
                        attr = "{}lim_{}".format(axis, side)
                        data_lim = func(
                            func(getattr(line, "get_{}data".format(axis))())
                            for line in ax.get_lines()
                        )
                        if scale != last_scale:
                            if hasattr(ax, attr):
                                delattr(ax, attr)
                        ext_lim = getattr(ax, attr, data_lim)
                        new_lim = func(ext_lim, data_lim)
                        if np.isfinite(new_lim) and not (
                            "log" in scale
                            and (np.isclose(new_lim, 0) or new_lim <= 0)
                        ):
                            setattr(ax, attr, func(new_lim, data_lim))
                            set_lim_kwargs[lim[axis][side]] = new_lim
                    if (
                        not np.isclose(*(set_lim_kwargs.values()))
                        if len(set_lim_kwargs) == 2
                        else True
                    ):
                        set_lim(**set_lim_kwargs)
                    setattr(ax, last_scale_attr, scale)
            else:
                for attr in filter(
                    functools.partial(hasattr, ax),
                    (
                        "{}lim_{}".format(a, s)
                        for a, s in itertools.product(("xy"), ("min", "max"))
                    ),
                ):
                    delattr(ax, attr)
                ax.autoscale(
                    enable=True, axis="both" if axes == "xy" else axes
                )
                ax.autoscale_view()



# system modules
import functools
import logging
from abc import ABC, abstractmethod

# internal modules

# external modules

logger = logging.getLogger(__name__)


class Filter(ABC):
    """
    The :class:`Filter` is a class to modify values before plotting them in the
    :any:`Animator`.
    """

    @property
    def when_parser(self):
        """
        Only operate on datasets from a given parser

        .. note::

            Set this option from the command-line via
            ``polt add-filter -o when-parser=PARSER``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_parser
        except AttributeError:
            self._when_parser = None
        return self._when_parser

    @when_parser.setter
    def when_parser(self, new):
        self._when_parser = str(new) if new else None

    @staticmethod
    def chain(*filters):
        """
        Chain filters together

        Args:
            filters (sequence of callables, optional): filters to chain

        Returns:
            callable : function taking a value as argument and returning the
            result of applying all given filters in order
        """
        return functools.partial(functools.reduce, lambda d, f: f(d), filters)

    @abstractmethod
    def update_dataset(self, dataset):
        """
        Modify a given dataset

        Args:
            dataset (dict): the dataset to modify

        Returns:
            dict or None: the modified dataset. ``None`` can be used to
            indicate that the dataset was modified in-place.

        .. note::

            This is an :any:`abc.abstractmethod`, so subclasses have to
            override this
        """

    def __call__(self, dataset):
        """
        If the :any:`when_parser` condition is not met just return given
        dataset. Otherwise, call :any:`update_dataset` with the given dataset.
        If that results in ``None``, assume that :any:`update_dataset` modified
        the dataset in-place and return it. Otherwise, return the return value
        of :any:`update_dataset`.
        """
        if self.when_parser is not None:
            if dataset.get("parser") != self.when_parser:
                return dataset
        modified = self.update_dataset(dataset)
        return dataset if modified is None else modified



# system modules

# internal modules
import polt.version
from polt.version import __version__
import polt.l10n
import polt.animator
import polt.extensions
import polt.streamer
import polt.parser
import polt.registry
import polt.utils
import polt.config
import polt.filter

# external modules



# system modules
import re
import logging
import io

# internal modules
from polt.l10n import _
from polt.parser.parser import Parser

# external modules

logger = logging.getLogger(__name__)


class NumberParser(Parser):
    """
    Simple parser extracting numbers line-wise

    Args:
        f (file-like object, optional): the connection to read from
        name (str, optional): human-readable description of the parser
        quantity (str, optional): quantity of the numbers read
        unit (str, optional): unit of the numbers read
        key (str, optional): key of the numbers read
    """

    NUMBER_REGEX = re.compile(
        r"-?\s*\d+(?:\.\d+)?(?:e-?\d+)?", flags=re.IGNORECASE
    )
    """
    Regular expression to detect numbers
    """

    def __init__(self, f=None, name=None, quantity=None, unit=None, key=None):
        Parser.__init__(self, f=f, name=name)
        if quantity is not None:
            self.quantity = quantity
        if unit is not None:
            self.unit = unit
        if key is not None:
            self.key = key

    @property
    def quantity(self):
        """
        The name of the quantity of the parsed numbers.

        .. note::

            Set this option from the command-line via
            ``polt add-source -p numbers -o quantity=NAME``

        """
        try:
            self._quantity
        except AttributeError:
            self._quantity = _("number")
        return self._quantity

    @quantity.setter
    def quantity(self, new):
        self._quantity = new

    @property
    def unit(self):
        """
        The unit of the parsed numbers.

        .. note::

            Set this option from the command-line via
            ``polt add-source -p numbers -o unit=UNIT``

        """
        try:
            self._unit
        except AttributeError:
            self._unit = None
        return self._unit

    @unit.setter
    def unit(self, new):
        self._unit = new

    @property
    def key(self):
        """
        The key of the parsed numbers.

        .. note::

            Set this option from the command-line via
            ``polt add-source -p numbers -o key=KEY``

        """
        try:
            self._key
        except AttributeError:
            self._key = None
        return self._key

    @key.setter
    def key(self, new):
        self._key = new

    @property
    def data(self):
        logger.debug(
            _("Reading from {}...").format(
                self.f.name if hasattr(self.f, "name") else repr(self.f)
            )
        )
        while True:
            try:
                line = next(self.f)
            except (UnicodeDecodeError, ValueError):
                continue
            except StopIteration:
                break
            if not line.strip():
                continue
            logger.debug(
                _("Received {} from {}").format(
                    repr(line),
                    self.f.name if hasattr(self.f, "name") else repr(self.f),
                )
            )
            numbers = (
                float(re.sub(r"\s*", r"", m.group()))
                for m in self.NUMBER_REGEX.finditer(
                    line.decode(errors="ignore")
                    if hasattr(line, "decode")
                    else line
                )
            )
            numbers_list = list(numbers)
            logger.debug(_("Numbers extracted: {}").format(numbers_list))
            yield {(self.quantity, self.unit, self.key): numbers_list}



# system modules
import logging
import itertools
import functools
import collections
import time
import sys
import csv
import random
import re
import math

# internal modules
import polt
from polt.l10n import _
from polt.cli.commands.main import cli

# external modules
import click

logger = logging.getLogger(__name__)

column_spec_regex = re.compile(
    r"(?:([^=(]*?)\s*=?\s*)?"  # name
    r"(?:"
    r"([^=(]+)"  # function
    r"(?:\s*\(\s*([^)]*)\s*\))?"  # options
    r")?"
)


def matching_random_function(spec, *args, **kwargs):
    """
    Generator yielding the next matching random number generating method of the
    :class:`random.Random` class. A method is considered a matching random
    number generator if its name contains the given specification, it returns a
    single :class:`float` and produces different values when called multiple
    times.

    Args:
        spec (str): the function specification string
        args, kwargs: further arguments to hand to the function

    Yields:
        callable : the matching random number generator wrapped with
            :func:`functools.partial` callable without arguments
    """
    for attr in filter(lambda x: spec in x, reversed(dir(random.Random))):
        obj = getattr(random, attr, None)
        if not hasattr(obj, "__call__"):
            continue
        func = functools.partial(obj, *args, **kwargs)
        try:
            assert isinstance(func(), float), "not float return value"
            assert len(set(func() for i in range(10))) == 10, "not random"
        except BaseException as e:
            continue
        yield func


def column_spec(ctx, param, value):
    m = column_spec_regex.fullmatch(value.strip())
    if not m:
        raise click.BadParameter(
            _("Specify a column like {examples}, not like {given}").format(
                examples=polt.l10n.join(
                    map(
                        repr,
                        map(
                            str.upper,
                            itertools.accumulate(
                                (
                                    _("name"),
                                    "=" + _("function"),
                                    "({})".format(_("options")),
                                )
                            ),
                        ),
                    ),
                    last_sep=" {} ".format(_("or")),
                ),
                given=repr(value),
            ),
            param=param,
            ctx=ctx,
        )
    col, func_spec, args_spec = m.groups()
    col = col if col else ("column" if func_spec is None else func_spec)
    args = tuple(
        map(
            functools.partial(polt.utils.to_float, default_on_failure=False),
            map(str.strip, args_spec.split(",")),
        )
        if args_spec
        else tuple()
    )
    default_funcs = {
        "sin": lambda: math.sin(
            time.time() * (args[0] if args else 1) * 2 * math.pi
        ),
        "cos": lambda: math.cos(
            time.time() * (args[0] if args else 1) * 2 * math.pi
        ),
        "time": time.time,
        "row": itertools.count(1).__next__,
        "walk": itertools.accumulate(
            random.uniform(-0.5, 0.5) for i in itertools.count()
        ).__next__,
    }
    if func_spec:
        func = default_funcs.get(func_spec)
        if not func:
            func = next(matching_random_function(func_spec, *args), None)
            if not func:
                raise click.BadParameter(
                    _(
                        "No random number generating method of random.Random "
                        "found having {spec} in its name and "
                        "taking {n_args} arguments {args}"
                    ).format(
                        spec=repr(func_spec),
                        n_args=len(args),
                        args="({})".format(
                            polt.l10n.join(
                                map(repr, args),
                                last_sep=" {} ".format(_("and")),
                            )
                        )
                        if args
                        else "",
                    ),
                    param=param,
                    ctx=ctx,
                )
    else:
        func = default_funcs.get(col, random.random)
    return (col, func)


@cli.command(
    help="\n\n".join(
        (
            _("generate random CSV data"),
            _("This command can be used to generate random CSV output."),
        )
    ),
    short_help=_("generate random CSV data"),
)
@click.option(
    "-o",
    "--output",
    metavar=_("file").upper(),
    help=_("File to write to. Defaults to standard output."),
    default=sys.stdout,
    type=click.File(mode="w"),
)
@click.option(
    "-c",
    "--column",
    "columns",
    metavar="[{}=]{}[({})]".format(
        _("name"), _("function"), _("options")
    ).upper(),
    help=_(
        "Add a column named {name_part} (or {func_part} unless specified). "
        "{func_part} may be (part of) the name "
        "of a random number generation method of the random.Random class "
        "(see https://docs.python.org/3/library/random.html) optionally "
        "taking the options {options_part} (comma-separated list of values). "
        "Specifying 'time' for {func_part} "
        "adds a column containing the current time as floating-point epoch "
        "timestamp. Equally, 'row' may be used to add a column containing "
        "the index of the current row. 'walk' generates a random walk. "
        "'sin(Hz)' or 'cos(Hz)' can be used to create waves. "
        "The default {func_part} is a uniform "
        "distribution in the interval [0, 1)."
    ).format(
        name_part=_("name").upper(),
        func_part=_("function").upper(),
        options_part=_("options").upper(),
    ),
    multiple=True,
    callback=lambda c, p, v: collections.OrderedDict(
        map(functools.partial(column_spec, c, p), v)
    ),
)
@click.option(
    "--header/--no-header",
    "output_header",
    help=_("Whether to output the header"),
    default=True,
    is_flag=True,
)
@click.option(
    "-d",
    "--delimiter",
    metavar=_("character").upper(),
    help=_("The delimiter to use. The default is a comma (,)."),
    default=",",
)
@click.option(
    "-r",
    "--max-rate",
    metavar=_("Hertz").upper(),
    help=_("Limit the data output rate (rows per second)"),
    default=float("inf"),
    type=click.FloatRange(min=sys.float_info.min),
)
@click.option(
    "-n",
    "--rows",
    metavar=_("amount").upper(),
    help=_("Limit the number of output data rows"),
    type=click.IntRange(min=0),
)
@click.option(
    "--stop-after",
    metavar=_("seconds").upper(),
    help=_("Stop outputting data after a given amount of seconds."),
    default=float("inf"),
    type=click.FloatRange(min=sys.float_info.min),
)
@click.option(
    "-s",
    "--seed",
    metavar=_("integer").upper(),
    help=_("Seed to use"),
    default=float("inf"),
    type=click.FLOAT,
)
@click.option(
    "-p",
    "--precision",
    metavar=_("integer"),
    help=_("How many precision digits to diplay for all columns"),
    type=click.IntRange(min=0),
)
def generate(
    columns,
    output,
    output_header,
    delimiter,
    max_rate,
    stop_after,
    rows,
    seed,
    precision,
):
    random.seed(seed)
    columns = columns or collections.OrderedDict((("random", random.random),))
    start_time = time.time()
    min_tdiff = 1 / max_rate
    last_output_time = float("-inf")
    writer = csv.DictWriter(
        output, fieldnames=columns.keys(), delimiter=delimiter
    )
    if output_header:
        writer.writeheader()
    formatter = (
        str if precision is None else "{{:.{}f}}".format(precision).format
    )
    for i in itertools.count() if rows is None else range(rows):
        if time.time() > start_time + stop_after:
            break
        if time.time() - last_output_time < min_tdiff:
            continue
        writer.writerow(
            {col: formatter(func()) for col, func in columns.items()}
        )
        output.flush()
        last_output_time = time.time()



# system modules
import re
import logging
import sys
import re
import functools
import itertools
import csv
import collections
import io


# internal modules
from polt.l10n import _
from polt.parser.parser import Parser
from polt.utils import str2num

# external modules

logger = logging.getLogger(__name__)


class CsvParser(Parser):
    """
    Parser for CSV input

    Args:
        f (file-like object, optional): the connection to read from
        name (str, optional): human-readable description of the parser
        header_regex (re.compile, optional): regular expression applied to the
            header to extract quantity and unit. See :attr:`header_regex` and
            :meth:`parse_header` for further information.
        only_columns (set of str, optional): only return columns in this
            set. The default is to use all columns without restriction.
        not_columns (set of str, optional): don't return columns in this
            set. This condition is stronger than ``only_columns``. The default
            is to use all columns without restriction.
        delimiter (str, optional): the delimiter to use. Default is a comma
            (``,``).
    """

    HEADER_REGEX_MATCHALL = re.compile(".*")
    """
    Default regular expression for :attr:`header_regex` matching everything.
    """

    HEADER_REGEX_UNIT_LAST = re.compile(
        pattern=r"(?P<quantity>.*?)(?:[^a-zA-Z0-9]+(?P<unit>[a-zA-Z0-9]+))?"
    )
    """
    Regular expression for use in :attr:`header_regex` or :meth:`parse_header`
    which extracts the unit as the last part of the header separated by a
    non-word character.
    """

    HEADER_REGEX_KEY_QUANTITY_UNIT = re.compile(
        pattern=r"(?P<key>[a-zA-Z0-9]+)"
        r"(?:"
        r"[^a-zA-Z0-9]+(?P<quantity>[a-zA-Z0-9]+)"
        r"(?:"
        r"[^a-zA-Z0-9]+(?P<unit>[a-zA-Z0-9]+)"
        r")?"
        r")?"
    )
    """
    Regular expression for use in :attr:`header_regex` or :meth:`parse_header`
    which extracts the key, then optionally the quantity and then optionally
    the unit from the header separated by a non-word character.
    """

    def __init__(
        self,
        f=None,
        name=None,
        header_regex=None,
        only_columns=None,
        not_columns=None,
        delimiter=None,
        **dictreader_kwargs,
    ):
        Parser.__init__(self, f=f, name=name)
        self.dictreader_kwargs = dictreader_kwargs
        if header_regex is not None:
            self.header_regex = header_regex
        if only_columns is not None:
            self.only_columns = only_columns
        if not_columns is not None:
            self.not_columns = not_columns
        if delimiter is not None:
            self.delimiter = delimiter

    @property
    def header_regex(self):
        """
        Regular expression applied used as default in :meth:`parse_header`.
        Defaults to :attr:`HEADER_REGEX_MATCHALL`.

        :type: regular expression created with :meth:`re.compile`
        :setter: convert the given string to a pattern via :meth:`re.compile`.
            However, first it is tried to look for a class property from which
            the name looke similar to the given string. If a match is found,
            that regex is used. For example one could set :attr:`header_regex`
            to ``"unit-last"`` and :attr:`HEADER_REGEX_UNIT_LAST` would be
            used.
        """
        try:
            self._header_regex
        except AttributeError:
            self._header_regex = self.HEADER_REGEX_MATCHALL
        return self._header_regex

    @header_regex.setter
    def header_regex(self, new):
        if isinstance(new, str):
            for attr, value in self.__class__.__dict__.items():
                m = re.fullmatch(r"HEADER_REGEX_(\w+)", attr)
                if not m:
                    continue
                key = next(iter(m.groups()), None)
                if not key:
                    continue
                if (
                    key.lower().strip()
                    == re.sub(r"\W+", "_", new).lower().strip()
                ):
                    self._header_regex = value
                    return
        self._header_regex = re.compile(pattern=new)

    def modifies_csvreader(decorated_function):
        """
        Decorator for methods that require a reset of the :attr:`csvreader`
        """

        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            if hasattr(self, "_csvreader"):
                logger.debug(
                    _("Resetting csvreader before {}").format(
                        decorated_function
                    )
                )
                del self._csvreader  # pragma: no cover
            return decorated_function(self, *args, **kwargs)

        return wrapper

    @property
    def csvreader(self):
        """
        The underlying csv parser

        :type: :class:`csv.DictReader`
        """
        try:
            self._csvreader
        except AttributeError:
            if self.comment:
                # filter out commented lines
                pattern = re.compile(str(self.comment), flags=re.IGNORECASE)
                fh = (L for L in self.f if not pattern.search(L))
            else:
                fh = self.f
            self._csvreader = csv.DictReader(fh, **self.dictreader_kwargs)
        return self._csvreader

    @property
    def dictreader_kwargs(self):
        """
        Further keyword arguments for the underlying :attr:`csvreader`

        :type: :class:`dict`
        :setter: resets the :attr:`csvreader`
        """
        try:
            self._dictreader_kwargs
        except AttributeError:  # pragma: no cover
            self._dictreader_kwargs = {}
        return self._dictreader_kwargs

    @dictreader_kwargs.setter
    @modifies_csvreader
    def dictreader_kwargs(self, new):
        self._dictreader_kwargs = new

    @Parser.f.setter
    @modifies_csvreader
    def f(self, new):
        if hasattr(new, "mode"):
            if "b" in new.mode:
                logger.debug(
                    _(
                        "csv.DictReader cannot handle bytes input, "
                        "using utf-8 wrappter"
                    )
                )
                new = io.TextIOWrapper(new, "utf-8")
        self._f = new

    @property
    def only_columns(self):
        """
        Only yield these columns in :attr:`data`. The default is to use all
        columns.

        .. note::

            Set this option from the command-line via
            ``polt add-source -p csv -o only-columns=column1,column2,...``

        :type: :class:`set`
        :setter: splits string inputs by commas and converts input to
            :class:`set`
        """
        try:
            self._only_columns
        except AttributeError:
            self._only_columns = None
        return self._only_columns

    @only_columns.setter
    def only_columns(self, new):
        self._only_columns = set(
            new.split(",") if hasattr(new, "split") else new
        )

    @property
    def not_columns(self):
        """
        Don't yield these columns in :attr:`data`. This condition is stronger
        than :attr:`only_columns`. The default is to use all columns.

        .. note::

            Set this option from the command-line via
            ``polt add-source -p csv -o not-columns=column1,column2,...``

        :type: :class:`set`
        :setter: splits string inputs by commas and converts input to
            :class:`set`
        """
        try:
            self._not_columns
        except AttributeError:
            self._not_columns = set()
        return self._not_columns

    @not_columns.setter
    def not_columns(self, new):
        self._not_columns = set(
            new.split(",") if hasattr(new, "split") else new
        )

    @property
    def delimiter(self):
        """
        Delimiter to use. This is a convenience property to allow for
        simple setup via the command-line and configuration.

        .. note::

            Set this option from the command-line via
            ``polt add-source -p csv -o delimiter=DELIMITER``

        :type: :class:`str`
        :getter: returns the ``delimiter`` from :attr:`dictreader_kwargs`
        :setter: sets the ``delimiter`` (remove if ``None``) and sets
            :attr:`dictreader_kwargs`
        """
        return self.dictreader_kwargs.get("delimiter")

    @delimiter.setter
    @modifies_csvreader
    def delimiter(self, new):
        d = self.dictreader_kwargs
        if new is None:
            d.pop("delimiter", None)
        else:
            d.update(delimiter=str(new))
        self.dictreader_kwargs = d

    @property
    def comment(self):
        """
        Regular expression for comment lines to ignore

        .. note::

            Set this option from the command-line e.g. via
            ``polt add-source -p csv -o comment='^#'`` to ignore lines starting
            with a ``#`` sign.

        :type: :class:`str`
        :getter: returns the current comment regex
        :setter: sets the comment regex
        """
        try:
            return self._comment
        except AttributeError:
            self._comment = None
        return self._comment

    @comment.setter
    @modifies_csvreader
    def comment(self, new):
        self._comment = new

    def str2num(self, s):
        """
        Convert a string or a sequence of strings to numbers with
        :meth:`polt.utils.str2num`.

        Args:
            s (str or sequence of str): the string(s) to convert

        Returns:
            (sequence of) str or int or float: the result of
                :meth:`polt.utils.str2num`
        """
        if isinstance(s, str):
            return str2num(s)
        elif isinstance(s, collections.abc.Sequence):
            return [self.str2num(x) for x in s]
        else:
            return s

    def parse_header(self, header, regex=None):
        """
        Parse a header field name based on :attr:`header_regex` to determine
        its quantity name, unit and key. The quantity is extracted as either
        the content of the ``quantity`` named captured group or the first
        captured group with a fallback to the given header name itself. The
        unit is extracted similarly as either the content of the ``unit`` named
        captured group or the next captured group. The same for the key. By
        default, the unit and key are both ``None``.

        Args:
            header (str): the header field name
            regex (re.compile, optional): the regular expression to use.
                Default is :attr:`header_regex`.

        Returns:
            quantity, unit, key: quantity, unit and key
        """
        m = (regex if regex is not None else self.header_regex).fullmatch(
            header
        )
        quantity = header
        unit = None
        key = None
        if m:
            sources = list(m.groups())
            d = m.groupdict().copy()
            for p in ("quantity", "unit", "key"):
                val = d.get(p)
                if val in sources:
                    sources.remove(val)
                if p not in d:
                    if sources:
                        v = sources.pop(0)
                        if v is not None:
                            d.setdefault(p, v)
            d.setdefault("quantity", header)
            quantity = d.get("quantity")
            unit = d.get("unit")
            key = d.get("key")
        return quantity, unit, key

    @property
    def data(self):
        while True:
            try:
                row = next(self.csvreader)
            except StopIteration:
                break
            except csv.Error as e:
                logger.error(f"csv parsing error: {e}")
                continue
            row = {
                (k.strip() if hasattr(k, "strip") else k): v
                for k, v in row.items()
            }
            d = {
                self.parse_header(k): self.str2num(v)
                for k, v in row.items()
                if (
                    k is not None
                    and k
                    in (
                        self.only_columns
                        if self.only_columns is not None
                        else set(row)
                    )
                    - self.not_columns
                )
            }
            if d:
                yield d



# system modules
import collections
import logging
import re
import importlib

# internal modules
from polt.filter import Filter
from polt import utils

# external modules
import numpy as np


logger = logging.getLogger(__name__)


class CalculationFilter(Filter):
    """
    Filter for modifying values with calculations
    """

    @property
    def when_quantity(self):
        """
        Only operate on quantities matching this regular expression

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o when-quantity=PATTERN``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_quantity
        except AttributeError:
            self._when_quantity = re.compile(".*")
        return self._when_quantity

    @when_quantity.setter
    def when_quantity(self, new):
        self._when_quantity = re.compile(
            str(new) if new else r".*", flags=re.IGNORECASE
        )

    @property
    def when_unit(self):
        """
        Only operate on units matching this regular expression

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o when-unit=PATTERN``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_unit
        except AttributeError:
            self._when_unit = re.compile(".*")
        return self._when_unit

    @when_unit.setter
    def when_unit(self, new):
        self._when_unit = re.compile(
            str(new) if new else r".*", flags=re.IGNORECASE
        )

    @property
    def when_key(self):
        """
        Only operate on keys matching this regular expression

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o when-key=PATTERN``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_key
        except AttributeError:
            self._when_key = re.compile(".*")
        return self._when_key

    @when_key.setter
    def when_key(self, new):
        self._when_key = re.compile(
            str(new) if new else r".*", flags=re.IGNORECASE
        )

    @property
    def count(self):
        """
        Operate on this many past values. Defaults to 1.

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o count=NUMBER``

        :type: :class:`str` or ``None``
        """
        try:
            self._count
        except AttributeError:
            self._count = 1
        return self._count

    @count.setter
    def count(self, new):
        self._count = max(1, int(new))

    @property
    def new_quantity(self):
        """
        Change the quantity of the newly created data. The strings
        ``{quantity}``, ``{key}``, ``{unit}`` of the current data and
        ``{function}`` are replaced in this string.

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o new-quantity='other {quantity}'``

        :type: :class:`str` or ``None``
        """
        try:
            return self._new_quantity
        except AttributeError:
            self._new_quantity = "{function}({quantity})"
        return self._new_quantity

    @new_quantity.setter
    def new_quantity(self, new):
        self._new_quantity = new if new else None

    @property
    def new_unit(self):
        """
        Change the unit of the newly created data. The strings
        ``{quantity}``, ``{key}``, ``{unit}`` of the current data and
        ``{function}`` are replaced in this string.

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o new-unit='not {unit} anymore'``

        :type: :class:`str` or ``None``
        """
        try:
            return self._new_unit
        except AttributeError:
            self._new_unit = "{unit}"
        return self._new_unit

    @new_unit.setter
    def new_unit(self, new):
        self._new_unit = new if new else None

    @property
    def new_key(self):
        """
        Change the key of the newly created data. The strings
        ``{quantity}``, ``{key}``, ``{unit}`` of the current data and
        ``{function}`` are replaced in this string.

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f calculate -o new-key='not {key} anymore'``

        :type: :class:`str` or ``None``
        """
        try:
            return self._new_key
        except AttributeError:
            self._new_key = "{key}"
        return self._new_key

    @new_key.setter
    def new_key(self, new):
        self._new_key = new if new else None

    @property
    def function(self):
        """
        .. note::

          Examples:

          .. code-block:: sh

            # running mean over last 10 values
            polt add-filter -f calculate -o function=numpy.mean -o count=10

            # running median over last 10 values
            polt add-filter -f calculate -o function=numpy.median -o count=10

            # scaling a specific quantity and changing the unit
            polt add-filter -f calculate \\
                -o when-quantity='temperature.*celsius' \\
                -o function=lambda x: x + 273.15  \\
                -o new-unit=Kelvin

            # using a function myfunction() in a separate Python file myfile.py
            polt add-filter -f calculate -o function=myfile.myfunction
        """
        try:
            return self._function
        except AttributeError:
            self._function = None
        return self._function

    @function.setter
    def function(self, new):
        """ """
        if m := re.fullmatch(
            r"(?P<modulename>.*?)\.(?P<objectname>[a-z]+)",
            str(new),
            flags=re.IGNORECASE,
        ):
            module = importlib.import_module(m.groupdict()["modulename"])
            self._function = getattr(
                module,
                m.groupdict()["objectname"],
            )
        else:
            try:
                # 😈 Dirty eval() 😈
                func = eval(new)
                if not hasattr(func, "__call__"):
                    raise ValueError()
            except Exception:
                raise ValueError(
                    f"Neither a Python function path like "
                    f"'package.submodule.functionname' "
                    f"nor a function definition: {new!r}"
                )
            self._function = func

    @property
    def cache(self):
        try:
            self._cache
        except AttributeError:
            self._cache = collections.defaultdict(
                lambda: collections.deque([], maxlen=self.count)
            )
        return self._cache

    def update_dataset(self, dataset):
        """
        ...

        Args:
            dataset (dict): the dataset to modify

        Returns:
            dict or None: the modified dataset

        """
        if not self.function:
            return dataset
        data = dataset.get("data", {})
        newdata = {}
        for key, val in data.items():
            logger.debug(f"{key = }, {val = }")
            keyiter = iter(utils.to_tuple(key))
            props = {
                p: next(keyiter, None) for p in ("quantity", "unit", "key")
            }
            regexes = {
                p: getattr(self, f"when_{p}", re.compile(r".*")) for p in props
            }
            matches = {p: regexes[p].search(v or "") for p, v in props.items()}
            if not all(matches.values()):
                continue
            try:
                self.cache[key].extend(val)
            except TypeError:
                self.cache[key].append(val)
            # construct new properties
            newprops = {}
            for p in props:
                template = getattr(self, f"new_{p}", f"{{{p}}}")
                try:
                    newprops[p] = template.format(
                        **dict(
                            **props,
                            function=getattr(
                                self.function,
                                "__name__",
                                f"function-{id(self.function)}",
                            ),
                        )
                    )
                except (ValueError, KeyError) as e:
                    logger.error(
                        f"Couldn't format {template = !r}: "
                        f"{type(e).__name__} {e}"
                    )
                    newprops[p] = template

            cachearray = np.asfarray(
                tuple(map(utils.to_float, self.cache[key]))
            )
            logger.debug(f"{cachearray = }")
            logger.debug(f"{self.function = }")
            try:
                newval = self.function(cachearray)
            except Exception as e:
                logger.error(
                    f"{type(e)} during "
                    f"{getattr(self.function,'__name__',str(self.function))}"
                    f"({cachearray}): {e}"
                )
                continue
            newdata[
                tuple(
                    (
                        newprops[k].format(**props)
                        if isinstance(newprops[k], str)
                        else newprops[k]
                    )
                    for k in ("quantity", "unit", "key")
                )
            ] = newval
        # add new data to dataset
        data.update(newdata)
        return dataset



__version__ = "1.1.0"

POLT_PARSER_ENTRY_POINT = "polt.parser"
"""
The entry point name for :class:`Parser` extensions
"""

POLT_ANIMATOR_ENTRY_POINT = "polt.animator"
"""
The entry point name for :class:`Animator` extensions
"""

POLT_FILTER_ENTRY_POINT = "polt.filter"
"""
The entry point name for :class:`Filter` extensions
"""



# system modules
import sys
import threading
import inspect
import collections
import functools
import logging
import time
import datetime

# internal modules
from polt.l10n import _

# external modules


logger = logging.getLogger(__name__)


class Streamer(threading.Thread):
    """
    Class to asynchronously stream and buffer data from a
    :class:`polt.parser.parser.Parser` in a :class:`threading.Thread`.

    Args:
        parser (polt.parser.)
        buffer (list-like, optional): A buffer for the data. Defaults To a
            :class:`collections.deque`.
        flush_after (float, optional): Amount of seconds to buffer recieved
            before extending the :attr:`buffer` with it. Default is 0.
        max_rate (float, optional): Maximum data rate (datasets per second)
            to append to :attr:`buffer`. Default is unlimited.
    """

    def __init__(
        self, parser=None, buffer=None, flush_after=None, max_rate=None
    ):
        threading.Thread.__init__(
            self, daemon=True, name="{}Thread".format(type(self).__name__)
        )
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @property
    def stopevent(self):
        """
        Event to stop the thread gracefully

        :type: :class:`threading.Event`
        """
        try:
            self._stopevent
        except AttributeError:  # pragma: no cover
            self._stopevent = threading.Event()
        return self._stopevent

    @property
    def parser(self):
        """
        The parser to buffer data from

        :type: :class:`polt.parser.parser.Parser`
        """
        try:
            self._parser
        except AttributeError:  # pragma: no cover
            self._parser = None
        return self._parser

    @parser.setter
    def parser(self, new_parser):
        try:
            iter(new_parser.data)
        except AttributeError:  # pragma: no cover
            raise ValueError(
                "{} object does not have a 'data' property".format(
                    type(new_parser).__name__
                )
            )
        except TypeError:  # pragma: no cover
            raise ValueError(
                "dataset property {} object is not iterable".format(
                    type(new_parser.data).__name__
                )
            )
        self._parser = new_parser

    @property
    def buffer(self):
        """
        Buffer for received data

        :type: list-like, i.e. something with :meth:`list.append` and
            :meth:`list.extend`-like methods
        :getter: Initializes an empty :class:`collections.deque`
            buffer if none is yet available
        """
        try:
            self._buffer
        except AttributeError:
            self._buffer = collections.deque()
        return self._buffer

    @buffer.setter
    def buffer(self, new_buffer):
        assert any(
            hasattr(new_buffer, attr) for attr in ("extend", "append")
        ), (
            "{} object does not neither have " "append() nor extend() method"
        ).format(
            type(new_buffer).__name__
        )
        self._buffer = new_buffer

    @property
    def flush_after(self):
        """
        Time frame to buffer recieved data before appending it to the
        :attr:`buffer`. Default is 0 seconds (immediately).

        :type: :class:`float` in seconds
        """
        try:
            self._flush_after
        except AttributeError:
            self._flush_after = 0
        return self._flush_after

    @flush_after.setter
    def flush_after(self, new):
        self._flush_after = abs(float(new))

    @property
    def max_rate(self):
        """
        Time frame to buffer recieved data before appending it to the
        :attr:`buffer`. Default is 0.1 seconds.

        :type: :class:`float` in datasets per second (Hz)
        """
        try:
            self._max_rate
        except AttributeError:
            self._max_rate = float("inf")
        return self._max_rate

    @max_rate.setter
    def max_rate(self, new):
        self._max_rate = abs(float(new)) if new is not None else None

    def run(self):
        """
        Continuously read data from :attr:`Streamer.parser` into
        :attr:`Streamer.buffer` by iterating over :attr:`Parser.data`. Also
        add a field ``time_recieved_utc`` to the dataset containing the value
        of :meth:`datetime.datetime.utcnow`.
        """
        logger.debug("Starting parser {}".format(repr(self.parser.name)))
        time_last_flush = float("-inf")
        time_recorded_last = float("-inf")
        buf = collections.deque()
        while True:
            if self.stopevent.is_set():
                logger.debug(_("stopevent is set, stop reading data"))
                break
            try:
                d = next(self.parser.data)
                logger.debug(
                    _("Received dataset {dataset} from {parser}").format(
                        dataset=repr(d), parser=repr(self.parser.name)
                    )
                )
                if (time.time() - time_recorded_last) > 1 / self.max_rate:
                    buf.append(
                        {
                            "time_recieved_utc": datetime.datetime.utcnow(),
                            "parser": self.parser.name,
                            "data": d,
                        }
                    )
                    time_recorded_last = time.time()
                if len(buf):
                    if (time.time() - time_last_flush) > self.flush_after:
                        self.buffer.extend(buf)
                        buf.clear()
                        time_last_flush = time.time()
            except StopIteration:
                logger.info(
                    _("No more {} data on {}").format(
                        type(self.parser).__name__, repr(self.parser.name)
                    )
                )
                break

    def stop(self):
        """
        Stop this thread gracefully, finishing just reading the current dataset
        """
        logger.debug(_("Will stop buffering soon"))
        self.stopevent.set()

    def __del__(self):
        """
        Deconstructor
        """
        pass



# system modules
import configparser
import itertools
import subprocess
import inspect
import importlib
import logging
import shlex
import os
import re
import io
import sys
import atexit


# internal modules
import polt
from polt.l10n import _
from polt.utils import normalize_cmd

# external modules
import xdgspec

USER_CONFIG_FILE = os.path.join(
    xdgspec.XDGPackageDirectory("XDG_CONFIG_HOME", "polt").path, "polt.conf"
)
LOCAL_CONFIG_FILES = (".polt.conf", "polt.conf")

DEFAULT_CONFIG_FILES = (USER_CONFIG_FILE,) + LOCAL_CONFIG_FILES

logger = logging.getLogger(__name__)


class SourceSectionError(ValueError):
    """
    Exception raised when the format of a :any:`Configuration.source_section`
    is invalid.
    """


class FilterSectionError(ValueError):
    """
    Exception raised when the format of a :any:`Configuration.filter_section`
    is invalid.
    """


class PlotSectionError(ValueError):
    """
    Exception raised when the format of the plot section is invalid.
    """


class Configuration(configparser.ConfigParser):
    """
    Class for configurations.
    """

    SOURCE_SECTION_PREFIX = "source"
    """
    Prefix for sections specifying a source
    """

    FILTER_SECTION_PREFIX = "filter"
    """
    Prefix for sections specifying a filter
    """

    def __init__(self):
        configparser.ConfigParser.__init__(self, interpolation=None)

    @property
    def source_section(self):
        """
        Generator yielding source sections

        Yields:
            configparser.SectionProxy: the next source section
        """
        for name, section in self.items():
            if name.startswith(self.SOURCE_SECTION_PREFIX):
                yield section

    @property
    def filter_section(self):
        """
        Generator yielding filter sections

        Yields:
            configparser.SectionProxy: the next filter section
        """
        for name, section in self.items():
            if name.startswith(self.FILTER_SECTION_PREFIX):
                yield section

    def matching_source_section(self, command=None, parser=None):
        """
        Generator yielding source sections with a similar specification

        Args:
            cmd (str, optional): the command to check for
            parser (str, optional): the parser to check for

        Yields:
            configparser.SectionProxy: the next matching source section
        """
        for section in self.source_section:
            command_matches = False
            if command is not None:
                this_cmd = section.get("command")
                if not this_cmd:  # pragma: no cover
                    continue
                if normalize_cmd(this_cmd) == normalize_cmd(command):
                    command_matches = True
            else:
                command_matches = True
            parser_matches = False
            if parser is not None:
                this_parser = section.get("parser")
                parser_matches = this_parser == parser
            else:
                parser_matches = True
            if command_matches and parser_matches:
                yield section

    @staticmethod
    def to_string(value):
        """
        Convert a value to a sensible configuration string. :class:`bool`
        objects are converted to ``"yes"`` and ``"no"``, :class:`str` objects
        are left as they are and the rest is converted to :class:`str`.

        Args:
            value (object): the value to convert

        Returns:
            str : the converted string value
        """
        if isinstance(value, str):
            return value
        if isinstance(value, bool):
            return "yes" if value else "no"
        return str(value)

    def add_section(self, name):
        """
        Add a new section with :any:`configparser.ConfigParser.add_section` but
        but if the name already exists, append something so that it doesn't
        exist.

        Args:
            name (str): the name of the new section

        Returns:
            str : the name actually used for the new section
        """
        nonexistant_name = next(
            filter(
                lambda x: x not in self,
                itertools.chain(
                    (name,),
                    (
                        " ".join(map(str, (name, _("nr."), nr)))
                        for nr in itertools.count(2)
                    ),
                ),
            )
        )
        configparser.ConfigParser.add_section(self, nonexistant_name)
        return nonexistant_name

    def update_option(self, section, key, value=None, default=None):
        """
        Update an option value

        Args:
            section (str): the name of the section
            key (str): the key
            value (object, optional): the new value. If ``None`` (the default)
                the value is left untouched or set to the value of ``default``
                if it is defined and the ``key`` doesn't exist.
                Is converted with :meth:`to_string`.
            default (object, optional): default fallback value if ``value`` is
                ``None``. Is converted with :meth:`to_string`.
        """
        if section not in self:
            self.add_section(section)
        sec = self[section]
        if value is None:
            if default is not None:
                if key not in sec:
                    sec[key] = self.to_string(default)
        else:
            sec[key] = self.to_string(value)

    def section_to_filter(self, section):
        """
        Set up a :any:`Filter` from a given section.

        Args:
            section (configparser.SectionProxy): the filter section

        Returns:
            parser.filter.Filter: the filter

        Raises:
            FilterSectionError : if the section format is wrong
        """
        filter_spec = section.get("filter")
        if filter_spec is None:
            raise FilterSectionError(
                _("Section {section} does not specify a 'filter'").format(
                    section=repr(section.name)
                )
            )
        parts = filter_spec.split(".")
        try:
            package, classname = ".".join(parts[:-1]), parts[-1]
            module = importlib.import_module(package)
            filter_cls = getattr(module, classname)
        except (IndexError, ImportError, AttributeError):
            raise SourceSectionError(
                _(
                    "filter specification {filter_spec} in "
                    "section {section} is not "
                    "an importable object"
                ).format(
                    filter_spec=repr(filter_spec), section=repr(section.name)
                )
            )
        if isinstance(filter_cls, type):
            if not issubclass(filter_cls, polt.filter.Filter):
                logger.warning(
                    _(
                        "Specified filter class {spec_class} is not a "
                        "subclass of {filter_class} and might not work."
                    ).format(
                        spec_class=filter_cls, filter_class=polt.filter.Filter
                    )
                )
            try:
                filter_obj = filter_cls()
            except BaseException as e:
                raise FilterSectionError(
                    _(
                        "Could not instantiate filter class {cls} from "
                        "section {section}: {err}"
                    ).format(
                        section=repr(section.name),
                        cls=repr(
                            ".".join(
                                (filter_cls.__module__, filter_cls.__name__)
                            )
                        ),
                        err=e,
                    )
                )
        else:
            filter_obj = filter_cls
        # set filter properties
        configured_filter_properties = {
            re.sub(r"\W+", "_", k.split(".")[1]): v
            for k, v in section.items()
            if (k.startswith("filter.") and k.split(".")[1])
        }
        for prop, val in configured_filter_properties.items():
            try:
                logger.debug(
                    _("Setting {filter}.{prop} to {val}").format(
                        prop=repr(prop), val=repr(val), filter=repr(filter_obj)
                    )
                )
                setattr(filter_obj, prop, val)
            except BaseException as e:
                logger.warning(
                    _(
                        "Could not set attribute {prop} to {val} "
                        "on filter {filter}: {err}"
                    ).format(
                        prop=repr(prop),
                        val=repr(val),
                        filter=repr(filter_obj),
                        err=repr(e),
                    )
                )
        if not hasattr(filter_obj, "__call__"):
            raise FilterSectionError(
                _("Filter object {filter} is not callable").format(
                    filter=filter_obj
                )
            )
        try:
            filter_obj({})
        except BaseException as e:
            raise FilterSectionError(
                _(
                    "Filter object {filter} can't modify an empty dataset: "
                    "{err}"
                ).format(filter=filter_obj, err=e)
            )
        return filter_obj

    def section_to_streamer(self, section):
        """
        Set up a :any:`Streamer` and :any:`Parser` from a given section. If the
        section defines a ``cmd``, that command will be launched with
        :class:`subprocess.Popen` and its :attr:`subprocess.Popen.stdout` will
        be used as :any:`Parser.f` input. Without ``cmd`` or if ``cmd`` is a
        dash (``-``), :attr:`sys.stdin` is used.

        Args:
            section (configparser.SectionProxy): the source section

        Returns:
            parser.streamer.Streamer: the set-up streamer with the set-up
                parser

        Raises:
            SourceSectionError : if the section format is wrong
        """
        parser_spec = section.get("parser")
        if parser_spec is None:
            raise SourceSectionError(
                _("Section {section} does not specify a 'parser'").format(
                    repr(section.name)
                )
            )
        parts = parser_spec.split(".")
        try:
            package, classname = ".".join(parts[:-1]), parts[-1]
            module = importlib.import_module(package)
            parser_cls = getattr(module, classname)
        except (IndexError, ImportError, AttributeError):
            raise SourceSectionError(
                _(
                    "parser specification {parser_spec} in "
                    "section {section} is not "
                    "an importable class specification"
                ).format(
                    parser_spec=repr(parser_spec), section=repr(section.name)
                )
            )
        if not issubclass(parser_cls, polt.parser.parser.Parser):
            logger.warning(
                _(
                    "Specified parser class {spec_class} is not a subclass "
                    "of {parser_class} and might not work."
                ).format(
                    spec_class=parser_cls,
                    parser_class=polt.parser.parser.Parser,
                )
            )
        cmd = section.get("command")
        encoding = section.get("encoding")
        if cmd:
            if cmd == "-":  # stdin
                if encoding:
                    try:
                        f = io.TextIOWrapper(
                            sys.stdin.buffer,
                            encoding=encoding,
                            errors="ignore",
                        )
                    except BaseException as e:
                        logger.error(
                            _(
                                "Could not use given encoding {encoding}: "
                                "{err}. Continuing without changing encoding."
                            ).format(encoding=repr(encoding), err=e)
                        )
                        f = sys.stdin
                else:
                    f = sys.stdin
            else:  # a command
                try:
                    process = subprocess.Popen(
                        shlex.split(cmd),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.DEVNULL,
                        errors="ignore",
                    )
                    atexit.register(lambda *a, **kw: process.kill())
                except BaseException as e:
                    raise SourceSectionError(
                        _("Could not start command {cmd}: {err}").format(
                            cmd=repr(cmd), err=e
                        )
                    )
                try:
                    returncode = process.wait(timeout=0.5)
                    if returncode == 0:
                        logger.warning(
                            _("Command {cmd} is already done").format(
                                cmd=repr(cmd)
                            )
                        )
                    else:
                        raise SourceSectionError(
                            _(
                                "Running {cmd} did not work "
                                "(return code {code})"
                            ).format(cmd=repr(cmd), code=returncode)
                        )
                except subprocess.TimeoutExpired:  # still running
                    logger.info(_("Successfully started {}").format(repr(cmd)))
                if encoding:
                    try:
                        # for backwards-compatibility with Python < 3.6 we
                        # can't just use the encoding and errors arguments of
                        # Popen().  Instead, we use an io.TextIOWrapper here.
                        f = io.TextIOWrapper(
                            process.stdout, encoding=encoding, errors="ignore"
                        )
                    except BaseException as e:
                        logger.error(
                            _(
                                "Could not use encoding {encoding} "
                                "for output of {cmd}: {err}"
                            ).format(
                                encoding=repr(encoding), cmd=repr(cmd), err=e
                            )
                        )
                        f = process.stdout
                else:
                    f = process.stdout
        else:
            raise SourceSectionError(
                _("Section {} does not specify a 'command'").format(
                    repr(section.name)
                )
            )
        source_text = _("stdin") if cmd == "-" else repr(cmd)
        parser_name = section.get("name") or _("{parser} on {source}").format(
            parser=parser_spec,
            source={"stdin": _("standard input")}.get(
                source_text, source_text
            ),
        )
        # create up parser
        try:
            parser = parser_cls(f)
        except BaseException as e:
            raise SourceSectionError(
                _(
                    "Could not instantiate parser {name} "
                    "with class {cls}: {err}"
                ).format(
                    name=repr(parser_name),
                    cls=repr(
                        ".".join((parser_cls.__module__, parser_cls.__name__))
                    ),
                    err=e,
                )
            )
        # set parser properties
        configured_parser_properties = {
            re.sub(r"\W+", "_", k.split(".")[1]): v
            for k, v in section.items()
            if (k.startswith("parser.") and k.split(".")[1])
        }
        for prop, val in configured_parser_properties.items():
            try:
                logger.debug(
                    _("Setting {parser}.{prop} to {val}").format(
                        prop=repr(prop), val=repr(val), parser=repr(parser)
                    )
                )
                setattr(parser, prop, val)
            except BaseException as e:
                logger.warning(
                    _(
                        "Could not set attribute {prop} to {val} "
                        "on parser {parser}: {err}"
                    ).format(
                        prop=repr(prop),
                        val=repr(val),
                        parser=repr(parser),
                        err=repr(e),
                    )
                )
        # create the streamer
        try:
            streamer = polt.streamer.Streamer(parser=parser)
        except BaseException as e:
            raise SourceSectionError(
                _(
                    "Could not instantiate the streamer for parser "
                    "{parser}: {err}"
                ).format(err=e, parser=parser.name)
            )
        # set streamer properties
        configured_streamer_properties = {
            re.sub(r"\W+", "_", k.split(".")[1]): v
            for k, v in section.items()
            if (k.startswith("streamer.") and k.split(".")[1])
        }
        for prop, val in configured_streamer_properties.items():
            try:
                logger.debug(
                    _("Setting {streamer}.{prop} to {val}").format(
                        prop=repr(prop), val=repr(val), streamer=repr(streamer)
                    )
                )
                setattr(streamer, prop, val)
            except BaseException as e:
                logger.warning(
                    _(
                        "Could not set attribute {prop} to {val} "
                        "on streamer {streamer}: {err}"
                    ).format(
                        prop=repr(prop),
                        val=repr(val),
                        streamer=repr(streamer),
                        err=repr(e),
                    )
                )
        return streamer

    @property
    def streamer(self):
        """
        Generator yielding :any:`Streamer` instances of all
        sections yielded from :any:`source_section` created with
        :any:`section_to_streamer`.
        """
        return map(self.section_to_streamer, self.source_section)

    @property
    def filter(self):
        """
        Generator yielding :any:`Filter` instances of all sections yielded from
        :any:`filter_section` created with :any:`section_to_filter`.
        """
        return map(self.section_to_filter, self.filter_section)

    @property
    def animator(self):
        """
        Create an animator from the configuration.
        """
        self.update_option("plot", "animator")
        section = self["plot"]
        animator_spec = section.get("animator")
        if not animator_spec:
            raise PlotSectionError(
                _("plot section does not contain an 'animator'")
            )
        parts = animator_spec.split(".")
        try:
            package, classname = ".".join(parts[:-1]), parts[-1]
            module = importlib.import_module(package)
            animator_cls = getattr(module, classname)
        except (IndexError, ImportError, AttributeError):
            raise PlotSectionError(
                _(
                    "animator specification {animator_spec} is not "
                    "an importable class specification"
                ).format(animator_spec=repr(animator_spec))
            )
        if not issubclass(animator_cls, polt.animator.Animator):
            logger.warning(
                _(
                    "Specified animator class {spec_class} "
                    "is not a subclass "
                    "of {animator_class} and might not work."
                ).format(
                    spec_class=animator_cls,
                    animator_class=polt.animator.Animator,
                )
            )
        try:
            animator = animator_cls()
        except BaseException as e:
            raise PlotSectionError(
                _(
                    "Could not instantiate animator " "with class {cls}: {err}"
                ).format(
                    cls=repr(
                        ".".join(
                            (animator_cls.__module__, animator_cls.__name__)
                        )
                    ),
                    err=e,
                )
            )
        # set animator properties
        configured_animator_properties = {
            re.sub(r"\W+", "_", k.split(".")[1]): v
            for k, v in section.items()
            if (k.startswith("animator.") and k.split(".")[1])
        }
        for prop, val in configured_animator_properties.items():
            try:
                logger.debug(
                    _("Setting {animator}.{prop} to {val}").format(
                        prop=repr(prop), val=repr(val), animator=repr(animator)
                    )
                )
                setattr(animator, prop, val)
            except BaseException as e:
                logger.warning(
                    _(
                        "Could not set attribute {prop} to {val} "
                        "on animator {animator}: {err}"
                    ).format(
                        prop=repr(prop),
                        val=repr(val),
                        animator=repr(animator),
                        err=repr(e),
                    )
                )
        return animator



# system modules
import functools
import collections
import logging
import inspect
import datetime
import operator
import re

# internal modules
from polt.l10n import _
from polt.animator.lines import LinesAnimator
from polt.utils import *

# external modules
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.dates import date2num, num2date

logger = logging.getLogger(__name__)


class TimeLinesAnimator(LinesAnimator):
    """
    Animator to display recieved data as lines.

    Args:
        args, kwargs: further arguments handed to the
            :any:`LinesAnimator` constructor.
        extrapolate (bool, optional): whether to extrapolate values (constant).
            Default is no extrapolation.
        share_time_axis (str, optional): whether to share the time axes.
            Default is ``True``.
        time_frame (int, optional): displayed time frame in seconds. By default
            (``None``), all times are diplayed.
        time_frame_buffer_ratio (float, optional): percentage of the time frame
            to expand the time axis into the future.
        show_data_rate (bool, optional): Whether to display the displayed data
            rate in the legend label. Defaults to ``False``.
    """

    def __init__(
        self,
        *args,
        extrapolate=None,
        share_time_axis=None,
        time_frame=None,
        time_frame_buffer_ratio=None,
        show_data_rate=None,
        **kwargs
    ):
        LinesAnimator.__init__(self, *args, **kwargs)
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @property
    def extrapolate(self):
        """
        Whether to extrapolate values constantly

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o extrapolate=yes|no``

        :type: :class:`bool`
        """
        try:
            self._extrapolate
        except AttributeError:
            self._extrapolate = False
        return self._extrapolate

    @extrapolate.setter
    def extrapolate(self, new):
        self._extrapolate = new == "yes" if isinstance(new, str) else bool(new)

    @property
    def share_time_axis(self):
        """
        Whether to share all time axes

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o share-time-axis=yes|no``

        :type: :class:`bool`
        """
        try:
            self._share_time_axis
        except AttributeError:
            self._share_time_axis = True
        return self._share_time_axis

    @share_time_axis.setter
    def share_time_axis(self, new):
        self._share_time_axis = (
            new == "yes" if isinstance(new, str) else bool(new)
        )

    @property
    def time_frame(self):
        """
        Time frame to display in seconds. Data outside the displayed window
        will be dropped.

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o time-frame=SECONDS``

        :type: :class:`int` or ``None``
        """
        try:
            self._time_frame
        except AttributeError:
            self._time_frame = None
        return self._time_frame

    @time_frame.setter
    def time_frame(self, new):
        self._time_frame = int(new)

    @property
    def time_frame_buffer_ratio(self):
        """
        Percentage of the time frame to expand the time axis into the future.
        Default is ``0.5`` which means 50%.

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o time-frame-buffer-ratio=RATIO``

        :type: :class:`float`
        """
        try:
            self._time_frame_buffer_ratio
        except AttributeError:
            self._time_frame_buffer_ratio = 0.5
        return self._time_frame_buffer_ratio

    @time_frame_buffer_ratio.setter
    def time_frame_buffer_ratio(self, new):
        assert new, "time_frame_buffer_ratio must be a positive float"
        self._time_frame_buffer_ratio = abs(float(new))

    @property
    def show_data_rate(self):
        """
        Whether to show the displayed data rate in the legend label. The
        default is ``False``.

        .. note::

            Set this option from the command-line via
            ``polt live -a lines -o show-data-rate=yes|no``

        :type: :class:`bool`
        """
        try:
            self._show_data_rate
        except AttributeError:
            self._show_data_rate = False
        return self._show_data_rate

    @show_data_rate.setter
    def show_data_rate(self, new):
        self._show_data_rate = (
            new == "yes" if isinstance(new, str) else bool(new)
        )

    def add_axes(self, *args, **kwargs):
        """
        If :any:`share_time_axis` is ``True``, this method adds the
        ``sharex`` argument to :any:`SubPlotsAnimator.add_axes`, otherwise it
        behaves exactly the same
        """
        add_axes_kwargs = kwargs.copy()
        if self.share_time_axis:
            first_ax = next(iter(self.figure.axes), None)
            if first_ax:
                add_axes_kwargs.update(sharex=first_ax)
        return LinesAnimator.add_axes(self, *args, **add_axes_kwargs)

    def new_line_coords(self, values, dataset={}):
        time_recieved_utc = dataset.get(
            "time_recieved_utc", datetime.datetime.utcnow()
        )
        times = to_tuple(time_recieved_utc) * len(values)
        return times, values

    def update_line(self, line, values, dataset={}):
        time_recieved_utc = dataset.get(
            "time_recieved_utc", datetime.datetime.utcnow()
        )
        times = to_tuple(time_recieved_utc) * len(values)
        x, y = line.get_xdata(), line.get_ydata()
        if (
            line.last_extrapolated
            if hasattr(line, "last_extrapolated")
            else False
        ):
            # if last value is just extrapolated, drop it
            x = x[:-1]
            y = y[:-1]
        x = np.append(x, times)
        y = np.append(y, values)
        line.set_xdata(x)
        line.set_ydata(y)
        line.last_extrapolated = False

    @LinesAnimator.register_method("matching-axes-found")
    def set_units(self, ax, props={}):
        """
        Update the figure with a single dataset

        Args:
            ax (matplotlib.axes.Axes): the axes
            props (dict, optional): properties of the current dataset
        """
        unit = props.get("unit")
        if not hasattr(ax, "units"):
            ax.units = set()
        if unit not in ax.units:
            ax.units.add(unit)
            ax.set_ylabel(
                " ".join(
                    [
                        _("value"),
                        " ".join(
                            "[{}]".format(unit)
                            for unit in sorted(
                                u if u is not None else _("unitless")
                                for u in ax.units
                            )
                        ),
                    ]
                )
            )

    @LinesAnimator.register_method("line-created")
    def expand_time_axis_if_first_line(self, line, *args, **kwargs):
        ax = line.axes
        if not ax:
            return
        if len(ax.get_lines()) == 1:
            now = datetime.datetime.utcnow()
            time_frame = datetime.timedelta(
                seconds=int(self.time_frame) if self.time_frame else 10
            )
            if line.get_xdata().size > 1:
                ax.set_xlim(now)
            else:
                ax.set_xlim(now, now + time_frame)

    @LinesAnimator.register_method("axes-added")
    def autofmt_xdate(self, *args, **kwargs):
        """
        Pretty-format the xaxis dates with
        :meth:`matplotlib.figure.Figure.autofmt_xdate` after resetting the
        x-axis tick params on all axes.

        Args:
            args, kwargs: just accept further arguments that might come in from
                the registering
        """
        for ax in self.figure.axes:
            ax.tick_params("x", labelbottom=True, reset=True)
            self.set_axes_labels(ax, y=False)
        self.figure.autofmt_xdate()

    @LinesAnimator.register_method("key-pressed")
    def autofmt_xdate_key_shortcut(self, event):
        """
        Callback to call :any:`autofmt_xdate` if the pressed
        key was an ``x``.

        .. hint::

            This means, you can press ``x`` when using a
            :any:`TimeLinesAnimator` to improve the xaxis date format

        Args:
            event (matplotlib.backend_bases.KeyEvent, optional): the key event
        """
        if event.key == "x":
            self.autofmt_xdate()

    @LinesAnimator.register_method("axes-added")
    def set_axes_labels(self, ax, x=True, y=True):
        if x:
            ax.set_xlabel(
                "{} [{}]".format(_("time"), plt.rcParams["timezone"])
            )
        if y:
            ax.set_ylabel(_("value"))

    @LinesAnimator.register_method("after-each-frame")
    def update_data_rate_in_legends(self, *args, **kwargs):
        """
        If :any:`show_data_rate` is ``True``, update the displayed data rate in
        each line's legend.
        """
        now = datetime.datetime.utcnow()
        for ax in self.figure.axes:
            for line in ax.get_lines():
                if self.show_data_rate:
                    x = line.get_xdata()
                    if (
                        line.last_extrapolated
                        if hasattr(line, "last_extrapolated")
                        else False
                    ):
                        x = x[:-1]
                    tdiff = (max(x.max(), now) - x.min()).total_seconds()
                    rate = max(x.size - 1, 0) / tdiff if tdiff > 0 else 0
                    rate_val, rate_str = (
                        (1 / rate, _("sec/ds"))
                        if (0 < rate < 1)
                        else (rate, _("Hz"))
                    )
                    rate_text = "({:.1f} {})".format(rate_val, rate_str)
                    rate_text_pattern = r"\([^)]+?({}|{})\)".format(
                        re.escape(_("Hz")), re.escape(_("sec/ds"))
                    )
                    old_label = new_label = line.get_label()
                    if re.search(rate_text_pattern, old_label):
                        new_label = re.sub(
                            rate_text_pattern, rate_text, old_label
                        )
                    else:
                        new_label += " " + rate_text
                    if old_label != new_label:
                        line.set_label(new_label)
                        ax.legend()

    @LinesAnimator.register_method("after-each-frame")
    def drop_old_values(self, *args, **kwargs):
        """
        If a :any:`time_frame` is set, drop values that are too old from all
        lines.
        """
        for ax in self.figure.axes:
            for line in ax.get_lines():
                # drop outside values to the left
                if self.time_frame:
                    lower, upper = ax.get_xlim()
                    x, y = line.get_xdata(), line.get_ydata()
                    outside = date2num(x) < lower
                    outside_ind = np.where(outside)[0]
                    if outside_ind.size:
                        line.set_xdata(np.delete(x, outside_ind[:-1]))
                        line.set_ydata(np.delete(y, outside_ind[:-1]))

    @LinesAnimator.register_method("after-each-frame")
    def extrapolate_lines(self, *args, **kwargs):
        """
        If :any:`extrapolate` is ``True``, extrapolate all lines constantly.
        """
        now = datetime.datetime.utcnow()
        for ax in self.figure.axes:
            for line in ax.get_lines():
                if self.extrapolate:
                    x, y = line.get_xdata(), line.get_ydata()
                    if len(x) == 1:
                        x = np.append(x, now)
                        y = np.append(y, y[-1])
                    elif len(x) > 1:
                        if (
                            line.last_extrapolated
                            if hasattr(line, "last_extrapolated")
                            else False
                        ):
                            x[-1] = now
                        else:
                            x = np.append(x, now)
                            y = np.append(y, y[-1])
                    else:
                        continue
                    line.last_extrapolated = True
                    line.set_xdata(x)
                    line.set_ydata(y)

    @LinesAnimator.register_method("after-each-frame")
    def adjust_time_axis_limits(self, *args, **kwargs):
        now = datetime.datetime.utcnow()
        for ax in self.figure.axes:
            new_lower, new_upper = lower, upper = tuple(
                map(
                    lambda x: num2date(x)
                    .astimezone(datetime.timezone.utc)
                    .replace(tzinfo=None),
                    ax.get_xlim(),
                )
            )
            past_diff = now - lower
            future_diff = upper - now
            future_ratio = future_diff / past_diff
            if not 0 < future_ratio < self.time_frame_buffer_ratio:
                while not new_upper > now:
                    new_upper += past_diff * self.time_frame_buffer_ratio
            if self.time_frame is not None:
                new_lower = new_upper - datetime.timedelta(
                    seconds=int(self.time_frame)
                )
            if new_lower != lower or new_upper != upper:
                ax.set_xlim(*map(date2num, (new_lower, new_upper)))
            if self.share_time_axis:
                break

    @LinesAnimator.register_method("after-each-frame")
    def adjust_view(self, *args, **kwargs):
        """
        Adjust the view with :any:`LinesAnimator.fit_view` for the y-axis.
        """
        self.fit_view(axes="y")



# system modules
import logging
import importlib
import subprocess
import shlex
import multiprocessing
import sys
import io

# internal modules
import polt
from polt.l10n import _
from polt.cli.utils import options_spec, python_class_spec
from polt.cli.commands.main import cli

# external modules
import click

logger = logging.getLogger(__name__)

animator_aliases = polt.extensions.EntryPointExtensions.get(
    polt.version.POLT_ANIMATOR_ENTRY_POINT, aliases=True
)
short_animator_aliases = polt.extensions.EntryPointExtensions.short_aliases(
    animator_aliases
)


def animate_process(buf, animator):
    logger.debug(_("Creating Animator"))
    animator.buffer = buf
    logger.debug(_("Running Animator"))
    animator.run()
    logger.debug(_("Animator is done"))


@cli.command(
    help="\n\n".join(
        (
            _("plot live data"),
            _(
                "This command parses the current configuration, "
                "sets up all parsers with their respective sources "
                "and launches the interactive plot window."
            ),
            _(
                "The animation can be paused and resumed "
                "by hitting the spacebar."
            ),
        )
    ),
    short_help=_("plot live data"),
)
@click.option(
    "--soft-shutdown/--hard-shutdown",
    "soft_shutdown",
    help=_("Shut down input threads gracefully"),
    default=None,
    is_flag=True,
)
@click.option(
    "-n",
    "--no-plot",
    help=_("Don't actually do any live plotting, just set the configuration"),
    default=False,
    is_flag=True,
)
@click.option(
    "-a",
    "--animator",
    "animator_spec",
    metavar=_("animator").upper(),
    help=_(
        "animator class to use. "
        "This can be a built-in animator ({native_animators}) "
        "or a class specification to a subclass of Animator "
        "(like 'mypackage.MyAnimator')"
    ).format(
        native_animators=polt.l10n.join(
            short_animator_aliases, last_sep=" {} ".format(_("or"))
        )
    ),
    default="lines",
    callback=python_class_spec(animator_aliases),
)
@click.option(
    "-o",
    "--option",
    "animator_options",
    metavar="{option}={value}".format(
        option=_("option"), value=_("value")
    ).upper(),
    multiple=True,
    callback=options_spec,
    help=_(
        "further options for the animator. "
        "Multiple specifications of this option are possible. "
        "The animator's attribute (previously snake-cased) {option} "
        "will be set to the string {value}. If {value} is not given, {option} "
        "is set to {value_default}."
    ).format(
        option=_("option").upper(),
        value=_("value").upper(),
        value_default=repr(polt.utils.to_float(True)),
    ),
)
@click.pass_context
def live(ctx, soft_shutdown, no_plot, animator_spec, animator_options):
    config = ctx.obj["config"]
    config.update_option("plot", "soft-shutdown", soft_shutdown)
    config.update_option("plot", "animator", animator_spec)
    for option, value in animator_options.items():
        config.update_option("plot", "animator.{}".format(option), value)

    if no_plot:
        logger.info(_("Skip plotting as requested via command-line"))
        return

    try:
        streamers = list(config.streamer)
    except polt.config.SourceSectionError as e:
        raise click.ClickException(
            _(
                "Problem creating streamer "
                "with parser from configuration: {err}"
            ).format(err=e)
        )

    if not streamers:
        logger.info(
            " ".join(
                [
                    _("No sources specified."),
                    _("Falling back to reading numbers from stdin."),
                ]
            )
        )
        if sys.stdin.isatty():
            click.echo(
                _(
                    "You may now repeatedly type any numbers "
                    "and press <Enter>."
                )
            )
        streamers = [
            polt.streamer.Streamer(
                parser=polt.parser.numberparser.NumberParser(sys.stdin)
            )
        ]

    # load filters
    try:
        filters = list(config.filter)
    except polt.config.FilterSectionError as e:
        raise click.ClickException(
            _("Could not load filters from configuration: {err}").format(err=e)
        )

    # load animator
    try:
        animator = config.animator
    except polt.config.PlotSectionError as e:
        raise click.ClickException(
            _("Could not load animator from configuration: {err}").format(
                err=e
            )
        )

    # set filters on animator
    animator.filters = filters

    with multiprocessing.Manager() as manager:

        # create shared buffer
        buf = manager.list()
        # connect streamer to buffers
        for streamer in streamers:
            streamer.buffer = buf
        p = multiprocessing.Process(
            target=animate_process,
            daemon=True,
            kwargs={"buf": buf, "animator": animator},
            name=_("PlotProcess"),
        )
        try:
            logger.debug(_("Starting plot process"))
            p.start()
            logger.debug(_("plot process started"))
            logger.debug(_("starting parsing threads"))
            for streamer in streamers:
                logger.debug(
                    _("starting parsing thread {}").format(streamer.name)
                )
                streamer.start()
            logger.debug(_("parsing threads started"))
            p.join()
            logger.debug(_("Plot process exited"))
        except KeyboardInterrupt:
            logger.info(_("User wants to stop"))
        if config.getboolean("plot", "soft-shutdown", fallback=False):
            logger.info(_("shutting down gracefully"))
            try:
                for streamer in streamers:
                    streamer.stop()
                    logger.debug(
                        _("Waiting for thread {} to close").format(
                            streamer.name
                        )
                    )
                    streamer.join()
            except KeyboardInterrupt:
                logger.info(_("Okay, okay, stopping now..."))

    logger.debug(_("live plotting done"))



# system modules
import logging
import sys
import os

# internal modules
from polt.l10n import _
from polt.config import DEFAULT_CONFIG_FILES
import polt.config

# external modules
import click

logger = logging.getLogger(__name__)


@click.group(
    help="\n\n".join(
        (
            "{} - {}".format(
                _("polt").title(), _("Live Data Visualisation via Matplotlib")
            ),
            _(
                "Commands can be chained and will "
                "be executed in the specified order."
            ),
            _(
                "Further information for each command can be viewed "
                "by passing --help to it (e.g. polt live --help)."
            ),
        )
    ),
    context_settings={"help_option_names": ["-h", "--help"]},
    chain=True,
)
@click.option("-q", "--quiet", help=_("only show warnings"), is_flag=True)
@click.option("-v", "--verbose", help=_("verbose output"), is_flag=True)
@click.option(
    "--no-config",
    help=_("don't load any default configuration files"),
    is_flag=True,
)
@click.version_option(help=_("show version and exit"))
@click.pass_context
def cli(ctx, quiet, verbose, no_config):
    # Add running directory to PYTHONPATH
    sys.path.insert(0, os.path.realpath(os.getcwd()))
    # set up logging
    loglevel = logging.DEBUG if verbose else logging.INFO
    loglevel = logging.WARNING if quiet else loglevel
    logging.basicConfig(
        level=loglevel, format="[%(asctime)s] - %(levelname)-8s - %(message)s"
    )
    for n, l in logger.manager.loggerDict.items():
        if not n.startswith("polt"):
            l.propagate = False
    ctx.ensure_object(dict)
    ctx.obj["config"] = polt.config.Configuration()
    if no_config:
        logger.debug(_("skip reading default configuration files"))
    else:
        logger.debug(
            _("read default configuration files {files}").format(
                files=", ".join(map(repr, DEFAULT_CONFIG_FILES))
            )
        )
        ctx.obj["config"].read(DEFAULT_CONFIG_FILES)



# system modules

# internal modules
from polt.filter import Filter
from polt.utils import *

# external modules


class MetaDataFilter(Filter):
    """
    Filter for modifying metadata of a dataset
    """

    @property
    def when_quantity(self):
        """
        Only operate on this quantity

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f metadata -o when-quantity=QUANTITY``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_quantity
        except AttributeError:
            self._when_quantity = None
        return self._when_quantity

    @when_quantity.setter
    def when_quantity(self, new):
        self._when_quantity = str(new) if new else None

    @property
    def when_unit(self):
        """
        Only operate on this unit

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f metadata -o when-unit=UNIT``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_unit
        except AttributeError:
            self._when_unit = None
        return self._when_unit

    @when_unit.setter
    def when_unit(self, new):
        self._when_unit = str(new) if new else None

    @property
    def when_key(self):
        """
        Only operate on this key

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f metadata -o when-key=KEY``

        :type: :class:`str` or ``None``
        """
        try:
            self._when_key
        except AttributeError:
            self._when_key = None
        return self._when_key

    @when_key.setter
    def when_key(self, new):
        self._when_key = str(new) if new else None

    @property
    def set_quantity(self):
        """
        Set the quantity to this value

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f metadata -o set-quantity=QUANTITY``

        :type: :class:`str` or ``None``
        """
        try:
            self._set_quantity
        except AttributeError:
            self._set_quantity = None
        return self._set_quantity

    @set_quantity.setter
    def set_quantity(self, new):
        self._set_quantity = str(new) if new else None

    @property
    def set_unit(self):
        """
        Set the unit to this value

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f metadata -o set-unit=UNIT``

        :type: :class:`str` or ``None``
        """
        try:
            self._set_unit
        except AttributeError:
            self._set_unit = None
        return self._set_unit

    @set_unit.setter
    def set_unit(self, new):
        self._set_unit = str(new) if new else None

    @property
    def set_key(self):
        """
        Set the key to this value

        .. note::

            Set this option from the command-line via
            ``polt add-filter -f metadata -o set-key=KEY``

        :type: :class:`str` or ``None``
        """
        try:
            self._set_key
        except AttributeError:
            self._set_key = None
        return self._set_key

    @set_key.setter
    def set_key(self, new):
        self._set_key = str(new) if new else None

    def update_dataset(self, dataset):
        """
        Modify parts of a dataset that match the conditions
        :any:`when_quantity`, :any:`when_unit` and :any:`when_key` and set
        their metadata according to the specified settings :any:`set_quantity`,
        :any:`set_unit` and :any:`set_key`.

        Args:
            dataset (dict): the dataset to modify

        Returns:
            dict or None: the modified dataset

        """
        data = dataset.get("data", {})
        rename = {}
        for key, val in data.items():
            keyiter = iter(to_tuple(key))
            props = {
                p: next(keyiter, None) for p in ("quantity", "unit", "key")
            }
            conditions = {
                p: getattr(self, "when_{}".format(p), None)
                for p in ("quantity", "unit", "key")
                if getattr(self, "when_{}".format(p), None) is not None
            }
            # skip this part if any conditions is not met
            if any(props[p] != v for p, v in conditions.items()):
                continue
            # determine new properties
            setprops = {
                p: getattr(self, "set_{}".format(p), None)
                for p in ("quantity", "unit", "key")
                if getattr(self, "set_{}".format(p), None) is not None
            }
            newprops = props.copy()
            newprops.update(setprops)
            if props == newprops:
                continue
            # make sure the new properties are not yet in the data
            propstuple = tuple(
                newprops[p] for p in ("quantity", "unit", "key")
            )
            while propstuple in data or propstuple in rename.values():
                propstuple += (None,)
            # record the new properties for later renaming
            rename[key] = propstuple
        # do the actual renaming
        for old, new in rename.items():
            data[new] = data.pop(old)

        return dataset



# system modules
import locale
import itertools
import gettext
import os
import sys
import ctypes
from pkg_resources import resource_filename

# internal modules

# external modules


GETTEXT_DOMAIN = "polt"
LOCALEDIR = resource_filename("polt", "locale")

# set up localization
locale.setlocale(locale.LC_ALL, "")

# taken from https://stackoverflow.com/a/42577134
try:
    if hasattr(locale, "bindtextdomain"):
        libintl = locale
    elif os.name == "nt":
        libintl = ctypes.cdll.LoadLibrary("libintl-8.dll")
    elif sys.platform == "darwin":
        libintl = ctypes.cdll.LoadLibrary("libintl.dylib")

    libintl.bindtextdomain(GETTEXT_DOMAIN, LOCALEDIR)

except (OSError, AttributeError):
    pass

gettext.bindtextdomain(GETTEXT_DOMAIN)
gettext.textdomain(GETTEXT_DOMAIN)
gettext.install(GETTEXT_DOMAIN, localedir=LOCALEDIR)
try:
    _ = gettext.translation(GETTEXT_DOMAIN, localedir=LOCALEDIR).gettext
except FileNotFoundError:
    _ = gettext.NullTranslations().gettext


def join(strings, sep=", ", last_sep=" {} ".format(_("and"))):
    """
    Joins a sequence of strings with a separator and use another separator for
    the last one.

    Args:
        strings (iterable of str): the strings to join
        sep (str, optional): the separator to use except for the last element
        last_sep (str, optional): the separator to use for the last element

    Returns:
        str : the joined string
    """
    strings = list(strings)
    if not strings:
        return ""
    if len(strings) > 1:
        return last_sep.join((sep.join(strings[:-1]), strings[-1]))
    else:
        return next(iter(strings), str(strings))
